/**
 * Copyright Damiano Bolla, Cologna Veneta, Italy, All right reserved
 * Eclipse does not have any more copyright manager, so I need to edit by hand
 * GNU LESSER GENERAL PUBLIC LICENSE Version 3, 29 June 2007
 */
#include <IPPapp.h>
#include <ipp_sock.h>
#include <stringd.h>
#include <ipiobuf.h>
#include "strtoksep.h"

struct Stat
   {
   struct IPP_soc ippsoc;
   struct IppAddress31 toIpp;    // for servers this will be the reply to address
   pthread_t manager_thread;     // manages events
   int isServer;
   uint16 server_port;     // for clients this is where to connect, for servers where to bind
   uint16 to_client_port;  // for servers, this will be where to send data to
   int app_exit;
   };

static struct Stat stat;


/**
 * send the message to parrot
 * @return TRUE
 */
static int sendto_message(char *lineP)
   {
   struct IPP_soc *sockP = &stat.ippsoc;

   if ( stat.isServer )
      {
      if ( ! stat.to_client_port )
         {
         printf("sendto_message SERVER waiting for client port\n");
         return TRUE;
         }

      // need to connect to a destination to be able to send anything
      ippsoc_connect_A(&stat.ippsoc, &stat.toIpp, stat.to_client_port);
      }

   if ( ippsoc_write(sockP, lineP, strlen(lineP)+1 ) < 0 )
      printf("sendto_message error: %s\n",sockP->soc_errmsg);

   return TRUE;
   }

/**
 * the format is ping [ipp address] port
 * port can be zero, meaning ping the router and not a specific port
 */
static int parse_input_ping(char *lineP)
   {
   char *nextP;

   char *ippaP = strtok_sep(lineP,&nextP, ' ');
   char *protoP = strtok_sep(NULL,&nextP, ' ');
   char *portP = strtok_sep(NULL,&nextP, ' ');
   char *ttlP = strtok_sep(NULL,&nextP, ' ');

   if ( strlen(ippaP) < 1 )
      {
      printf("ping missing to '%s' \n",ippaP);
      return TRUE;
      }

   struct IPPS_hri_ping_req ping_req;
   memzero_obj(ping_req);

   ping_req.icmp_code=ICMP_CODE_PING;
   ippsoc_string_to_address(&ping_req.to_address, ippaP);

   if ( protoP )
      ping_req.to_protocol=atoi(protoP);

   if ( portP )
      ping_req.to_port=atoi(portP);

   if ( ttlP )
      ping_req.initial_ttl=atoi(ttlP);

   struct IPP_soc *sockP = &stat.ippsoc;

   ippsoc_write_icmp(sockP, &ping_req, sizeof(ping_req));

   printf("ping to '%s' proto=%d port=%d ttl=%d sent \n",ippaP,ping_req.to_protocol,ping_req.to_port,ping_req.initial_ttl);

   return TRUE;
   }

/**
 * @return TRUE if I need to continue parsing
 */
static int parse_input_line(char *lineP)
   {
   char *paramsP;

   if ( (paramsP=string_starts_with(lineP,"quit")))
      return FALSE;

   if ( (paramsP=string_starts_with(lineP,"stats")))
      return ippsoc_print_stats(&stat.ippsoc) == DONE;

   if ( (paramsP=string_starts_with(lineP,"ping ")))
      return parse_input_ping(paramsP);

   return sendto_message(lineP);
   }


/**
 * @return TRUE or FALSE if need to end
 * radioptp_set_on_receive( ipface_on_raptp_receive );
 */
static int main_loop_continue(void)
   {
   char abuffer[1000];

   char *lineP = fgets(abuffer, sizeof(abuffer), stdin);

   if ( ! lineP ) return FALSE;

   string_strip_crlf(lineP);

   printf("> %s\n",lineP);

   return parse_input_line(lineP);
   }

static void parse_sock_reply_message(struct IPP_soc *const sockP, struct Ipiobuf *bufP)
   {
   struct IPPS_hrp_reply *resP = ipiobuf_get_pullP(bufP);

   if ( resP->hrp_rescode )
      {
      // I only print error messages
      printf("sock message: code=%d %s\n",resP->hrp_rescode,resP->hrp_message);
      }
   }


static void parse_rh_data(struct IPP_soc *const sockP, struct Ipiobuf *bufP, struct IppChunkh16 *firstcP)
   {
   struct IPPS_udp_data_rhp *resP = ipiobuf_get_pullP(bufP);

   uint16 rhp_datalen=firstcP->chunk_len-sizeof(struct IppChunkh16);
   uint16 payload_len=rhp_datalen-sizeof(struct IPPS_udp_data_rhp);

   if ( stat.isServer )
      {
      stat.toIpp=resP->from_address;
      stat.to_client_port=resP->from_port;
      }

   printf("received data: payload_len=%d %s\n",payload_len,resP->data_array);
   }

static void parse_rh_icmp_ping(struct IPP_soc *const sockP, struct Ipiobuf *bufP, struct IPPS_rhi_ping_res *pingP)
   {
   char tbuf[100];

   ippsoc_address_to_string(tbuf,sizeof(tbuf),&pingP->from_address);
   printf("ping RES reason=%d from %s\n",pingP->ping_reason,tbuf);

   }

static void parse_rh_icmp_unreach(struct IPP_soc *const sockP, struct Ipiobuf *bufP, struct IPPS_rhi_unreach_res *unrP)
   {
   char tbuf[100];

   ippsoc_address_to_string(tbuf,sizeof(tbuf),&unrP->o_to_address);
   printf("unreach RES reason=%d from %s\n",unrP->unreach_reason,tbuf);

   }


static void parse_rh_icmp(struct IPP_soc *const sockP, struct Ipiobuf *bufP, struct IppChunkh16 *firstcP)
   {
   union IPPS_rh_icmp_all *allP = ipiobuf_get_pullP(bufP);

   switch ( allP->icmp_code )
      {
      case ICMP_CODE_PING:
         parse_rh_icmp_ping(sockP,bufP,&allP->ping_res);
         break;

      case ICMP_CODE_UNREACHABLE:
         parse_rh_icmp_unreach(sockP,bufP,&allP->unreach_res);
         break;

      default:
         printf("parse_rh_icmp: unsupported ICMP code=%d\n",allP->icmp_code);
      }
   }

static void parse_sock_reply(struct IPP_soc *const sockP)
   {
   struct Ipiobuf *const bufP = sockP->rx_bufP;

   if ( ! bufP )
      {
      printf("parse_sock_reply: missing bufP connection closed ?\n");
      // by closing the stdin I will wignal that the demo is terminated
      exit(0);
      return;
      }

   // this also reset the pull index
   ipiobuf_pull_packet_len(bufP);

   struct IppChunkh16 *firstcP=ipiobuf_pull_chunkh16(bufP);

   if ( ! firstcP )
      {
      printf("parse_sock_reply: missing chunk\n");
      return;
      }

   switch (firstcP->chunk_type)
      {
      case IPPHRchu_message_rh:
         parse_sock_reply_message(sockP, bufP);
         break;

      case IPPHRchu_udp_data_rh:
         parse_rh_data(sockP,bufP,firstcP);
         break;

      case IPPHRchu_icmp_rh:
         parse_rh_icmp(sockP,bufP,firstcP);
         break;

      default:
         printf("parse_sock_reply: unmapped chunk %d\n",firstcP->chunk_type);
         break;
      }
   }


void parse_params(int argc, char *argv[])
   {
   // default values if nothing given
   stat.server_port=IPPUDP_PARROT_PORT;
   ippsoc_string_to_address(&stat.toIpp,"127");

   if ( argc < 2 )
      {
      printf("%s [server | client] [to IPP address] [port]\n",argv[0]);
      printf("if no parameters it will be a client to local router parrot port\n");
      return;
      }

   if ( argc > 1 )
      stat.isServer = argv[1][0] == 's';

   if ( argc > 2 )
      ippsoc_string_to_address(&stat.toIpp,argv[2]);

   if ( argc > 3 )
      stat.server_port=atoi(argv[3]);
   }

/**
 * This uses UDP, so, there is not really a "connection", just setting ports where to send and receive data
 * the application accept TWO optional parameters
 * 1 the port to CONNECT to, that is to send the data
 * 2 the port to BIND to, that is the port where to receive data from
 * both ports are optional
 * by default the demo will connect to router parrot port and leave the default source port
 * eg for a two way talk
 * IPPapp server 127 1234
 * IPPapp client 127 1234
 */
int main(int argc, char *argv[])
   {
   memzero_obj(stat);
   char tbuf[100];

   parse_params(argc, argv);

   // this is to connect to the router itself and define what we will be talking with (beside sendto)
   ippsoc_init(&stat.ippsoc, ICHUNK_UDP, "127.0.0.1",2024,0);
   ippsoc_open(&stat.ippsoc, parse_sock_reply);

   printf("IPP %s\n",ippsoc_address_to_string(tbuf,sizeof(tbuf),&stat.toIpp));
   printf("PORT %d\n",stat.server_port);

   if ( stat.isServer )
      {
      printf("MODE SERVER\n");
      // Server need to bind to the port where the client will send data
      // to reply to the incoming client packet, the server need to extract the FROM port and reply to that
      ippsoc_bind(&stat.ippsoc, &stat.toIpp, stat.server_port );
      }
   else
      {
      printf("MODE CLIENT\n");
      ippsoc_connect_A(&stat.ippsoc, &stat.toIpp, stat.server_port );
      }

   printf("type quit and then ENTER to exit\n");

   while (main_loop_continue());

   ippsoc_close(&stat.ippsoc);

   return EXIT_SUCCESS;
   }
