/**
 * Copyright Damiano Bolla, Cologna Veneta, Italy, All right reserved
 * Eclipse does not have any more copyright manager, so I need to edit by hand
 * GNU LESSER GENERAL PUBLIC LICENSE Version 3, 29 June 2007
 */

#include <ipiobuf.h>


struct IpiobufStaT
   {
   uint32 ipiobuf_mcount;        // malloc_count
   uint32 ipiobuf_fcount;        // free_count
   uint32 ipiobuf_excess_free;   // excessive, incorrect free
   };

static struct IpiobufStaT ipiobufStat;

/**
 * Cleaning up of the generic handler structure
 */
int ipiobuf_init(void)
   {
   memzero_obj(ipiobufStat);

   return DONE;
   }


void ipiobuf_print_stats(void)
   {
   CONSOLE_PRINTF("ipiobuf_mcount..... = %d\n",ipiobufStat.ipiobuf_mcount);
   CONSOLE_PRINTF("ipiobuf_fcount..... = %d\n",ipiobufStat.ipiobuf_fcount);
   CONSOLE_PRINTF("ipiobuf_excess_free = %d\n",ipiobufStat.ipiobuf_excess_free);
   }

/**
 * I cannot allocate more than uint16, by design
 * it will return a malloc pointer with data cleared to zero
 * the size is multiple of 1024, so the memory manager can avoid fragmentation
 *
 * I can PUSH data into the buffer
 * I can PULL data out of the buffer
 */
struct Ipiobuf * ipiobuf_malloc(uint8 iobuf_type, uint16 want_max_size)
   {
   // this is safety check that the compiler does not screw me up
   // the size of the structure MUST be exactly the one I calculate by hand
   if ( sizeof(struct Ipiobuf) != IPIOBUF_HEADER_LEN )
      {
      CONSOLE_PRINTF("ipiobuf_malloc: FATAL incorrect header len\n");
      return NULL;
      }

   uint16 w_size = want_max_size+sizeof(struct Ipiobuf);
   uint16 mult=w_size / IPIOBUF_MULTIPLEN;

   // this may really fail because no memory
   struct Ipiobuf * risulP = calloc(mult+1,IPIOBUF_MULTIPLEN);

   if ( ! risulP ) return NULL;

   uint16 h_max_size=(mult+1)*IPIOBUF_MULTIPLEN;

   ipiobufStat.ipiobuf_mcount++;

   risulP->object_type=iobuf_type;
   risulP->data_array_len=h_max_size;
   risulP->object_refcount=1;

   return risulP;
   }

/**
 * Attempt to release the iobuf
 * if the usage is zero it will free it
 * @returns NULL so you can use a single call to clean a reference
 */
struct Ipiobuf * ipiobuf_release ( struct Ipiobuf * const iobufP )
   {
   if ( ! iobufP )
      return NULL;

   // it means that someone has already free it and I am attempting to free it again
   // note that refcount is an uint8 so, if it is not zero it is some kind of number > 0
   if ( iobufP->object_refcount == 0 )
      {
      fprintf(stderr,"ipiobuf_release: WARNING double release\n");
      ipiobufStat.ipiobuf_excess_free++;
      return NULL;
      }

   iobufP->object_refcount--;

   // only if I am at zero, now, I can free it
   if ( iobufP->object_refcount == 0 )
      {
      free(iobufP);
      ipiobufStat.ipiobuf_fcount++;
      }

   return NULL;
   }

/**
 * NOTE the difference is between the push_index and the pull_index
 * @return the max number of chars that can be TX using a write syscall
 */
int ipiobuf_get_pull_max(struct Ipiobuf *  iobufP)
   {
   // if no iobuf is given, nothing to do
   if ( ! iobufP )  return 0;

   int delta = iobufP->push_index - iobufP->pull_index;

   return delta;
   }


/**
 * used to know if there is data that can be extracted from the iobuf
 * @return TRUE if something can be pulled from iobuf
 */
int ipiobuf_can_pull_something (struct Ipiobuf *  iobufP )
   {
   if ( ! iobufP ) return FALSE;

   return ipiobuf_get_pull_max(iobufP) > 0;
   }

/**
 * @return TRUE if the given number of bytes can be pulled
 */
int ipiobuf_can_pull(struct Ipiobuf *  iobufP, uint16 n_bytes )
   {
   if ( ! iobufP ) return FALSE;

   return ipiobuf_get_pull_max(iobufP) >= n_bytes;
   }


/**
 * This is NOT used when reading from the network, there is a special for that
 * @return the max number of chars that can be PUSH into the buffer
 */
static int ipiobuf_get_push_max(struct Ipiobuf * const iobufP)
   {
   int delta = iobufP->data_array_len - iobufP->push_index;

   return delta;
   }

/**
 * @return TRUE if it is possible to push the given number of bytes
 */
int ipiobuf_can_push(struct Ipiobuf * const iobufP, uint16 n_bytes)
   {
   return ipiobuf_get_push_max(iobufP) >= n_bytes;
   }

/**
 * Used to know if something else can be added
 */
int ipiobuf_can_push_something(struct Ipiobuf * iobufP)
   {
   return iobufP->data_array_len > iobufP->push_index;
   }

/**
 * You are reading from the network and wish to know if you have received everything
 * the test is against the packet_len that you MUST have already received (first two bytes of packet)
 * @return TRUE if the number of bytes received is >= the packet length
 */
int ipiobuf_rxnet_is_received(struct Ipiobuf * const iobufP)
   {
   return iobufP->push_index >= ipiobuf_get_packet_len(iobufP);
   }

/**
 * @return a pointer where you can further push data
 */
void *ipiobuf_get_pushP(struct Ipiobuf *  iobufP)
   {
   return iobufP->data_array + iobufP->push_index;
   }

/**
 * a get operation will not change the pull_index
 * @return the current pull pointer
 */
void *ipiobuf_get_pullP(struct Ipiobuf *  iobufP)
   {
   return iobufP->data_array+iobufP->pull_index;
   }


/**
 * This sets the chunk len as the difference between current insert pointer and the given location of the len byte
 * @return the calculated len or zero if errors
 */
uint8 ipiobuf_set_chulen8(struct Ipiobuf *  iobufP, uint8 *clenP )
   {
   void *lenP = clenP;

   int len = ipiobuf_get_pushP(iobufP) - lenP;

   if ( len > 0xff || len < 0 )
      return 0;

   len += 1;

   *clenP = len;

   return len;
   }

/**
 * this sets the chunk length as the difference between current insert pointer and the given location of the length uint16
 * @return the calculated length or zero if errors
 */
uint16 ipiobuf_set_chulen16(struct Ipiobuf * const iobufP, uint16 * const clenP )
   {
   void * vlenP = clenP;

   int len = ipiobuf_get_pushP(iobufP) - vlenP;

   if ( len > 0xFFFF || len < 0 )
      return 0;

   len += 2;

   *clenP = len;

   return len;
   }


/**
 * Sets the PACKET length as the current number of bytes written in the buffer
 * This must be called if you add something to be buffer
 * If instead, you wish to set the packet len of INCOMING data use the other one
 * @return the calculated length
 */
uint16 ipiobuf_set_pktlen_new(struct Ipiobuf *  ibufP )
   {
   if ( ! ibufP ) return 0;

   uint16 *lenP = (uint16 *)ibufP->data_array;

   *lenP = ibufP->push_index;

   return ibufP->push_index;
   }

/**
 * this is the one to be called ASAP when you have the two bytes that
 * declare the length of the following data
 * What it does is storing the incoming packet length and pushing the ins index past it
 * NOTE radio handling ALWAYS receive more than two bytes, so, you CANNOT use this one !!
 * NOTE that it may fail if for any reason there is no enough space
 */
uint16 ipiobuf_rxnet_push_packet_len(struct Ipiobuf * const ibufP, uint16 rx_pktlen )
   {
   if ( ! ibufP ) return 0;

   if ( rx_pktlen > ibufP->data_array_len ) return 0;

   uint16 *lenP = (uint16 *)ibufP->data_array;

   *lenP = rx_pktlen;

   ibufP->push_index = 2;

   return rx_pktlen;
   }

/**
 * This is used by radio receive to SET the packet length on the first blob of data
 * Radio receives a full blob at the beginning and the functions that stare away NEED to have the packet len set BEFORE writing !
 * the difference between set and push is that SET does NOT touch the push_index !
 */
uint16 ipiobuf_set_packet_len(struct Ipiobuf * const ibufP, uint16 pktlen )
   {
   if ( ! ibufP ) return 0;

   uint16 *lenP = (uint16 *)ibufP->data_array;

   *lenP = pktlen;

   return pktlen;
   }


/**
 * Gets the packet LENGTH, does NOT touch the pull index
 * So, you can SAFELY use this one to know how big the data packet should be
 */
uint16 ipiobuf_get_packet_len(struct Ipiobuf * const bufP)
   {
   if ( ! bufP ) return 0;

   uint16 *lenP = (uint16 *)bufP->data_array;

   return *lenP;
   }

/**
 * this pulls the packet length, meaning that it gets the length and advances the pull index
 * It is assumes that the length is at the beginning of the buffer !
 * This basically resets the pull index
 * TODO: in theory, the push index MUST be equal to the packet len
 * if not, it means that some part of data is missing !
 * @return the packet length as stored IN the buffer or zero if params invalid
 */
uint16 ipiobuf_pull_packet_len(struct Ipiobuf * const bufP)
   {
   if ( ! bufP ) return 0;

   uint16 *lenP = (uint16 *)bufP->data_array;

   bufP->pull_index = 2;

   return *lenP;
   }


/**
 * This is used when reading from network, rxnet
 * @return the max number of chars that can be rx, into the buffer
 */
int ipiobuf_rxnet_get_push_max(struct Ipiobuf *const iobufP)
   {
   int delta = ipiobuf_get_packet_len(iobufP) - iobufP->push_index;

   return delta;
   }



/**
 * Attempt to fill the iobuf with data from the given fd
 * This is the generic form where YOU should tell how much to fill up
 * @return the number of bytes received or zero if it fails
 */
int ipiobuf_rxnet_max (struct Ipiobuf *  iobufP, int unix_fd, int w_max )
   {
   if ( w_max <= 0 ) return 0;

   // the syscall read has to put some data into a WRITABLE space
   int letti = read(unix_fd, ipiobuf_get_pushP(iobufP), w_max);

   if ( letti <= 0 ) return 0;

   iobufP->push_index += letti;

   return letti;
   }

/**
 * Attempt to fill the iobuf with data from the given fd
 * NOTE that the first fill is just for two bytes, the length
 * and then no more than the given amount of data should be read
 * @return the number of bytes received
 */
int ipiobuf_rxnet_from (struct Ipiobuf *  iobufP, int unix_fd)
   {
   int rxmax = ipiobuf_rxnet_get_push_max(iobufP);

   return ipiobuf_rxnet_max(iobufP,unix_fd, rxmax);
   }


/**
 * If the index is odd, push one byte with value 1 in the buffer
 * @return 1 if pushed or zero if not
 */
int ipiobuf_push_even (struct Ipiobuf * iobufP )
   {
   if ( iobufP->push_index & 0x01 )
      {
      if ( ! ipiobuf_can_push(iobufP, 1)) return 0;
      iobufP->data_array[iobufP->push_index++] = ICHUNK_ONEBYTE;
      return 1;
      }

   return 0;
   }

/**
 * push an uint16 to a iobuf
 * for fast access this should be even aligned, so, if the index is odd an extra byte will be written
 * @return the number of bytes written or zero
 */
int ipiobuf_push_uint16 (struct Ipiobuf * iobufP, uint16 ashort)
   {
   ipiobuf_push_even(iobufP);

   if ( ! ipiobuf_can_push(iobufP, sizeof(ashort))) return 0;

   memcpy(iobufP->data_array+iobufP->push_index, &ashort, sizeof(ashort));

   iobufP->push_index += sizeof(ashort);

   return sizeof(ashort);
   }

/**
 * @return the number of bytes written or zero
 */
int ipiobuf_push_uint32 (struct Ipiobuf * iobufP, uint32 aval)
   {
   ipiobuf_push_even(iobufP);

   if ( ! ipiobuf_can_push(iobufP, sizeof(aval))) return 0;

   memcpy(iobufP->data_array+iobufP->push_index, &aval, sizeof(aval));

   iobufP->push_index += sizeof(aval);

   return sizeof(aval);
   }


/**
 * Align an iopbuf reading to even, once you have done with a chunk
 */
int ipiobuf_pull_even (struct Ipiobuf * iobufP)
   {
   if ( iobufP->pull_index & 0x01 )
      {
      // if there is not enough to pull
      if ( ! ipiobuf_can_pull(iobufP, 1)) return 0;
      iobufP->pull_index++;
      return 1;
      }

   return 0;
   }

/**
 * pull a uint16 from the iobuf
 * NOTE that uint16 (and possibly longer integers MUST be even aligned)
 * @return the uint8 value
 */
uint16 ipiobuf_pull_uint16 (struct Ipiobuf * iobufP)
   {
   uint16 ashort;

   ipiobuf_pull_even(iobufP);   // if the push has an even, the pull has it too

   // if there is not enough to pull
   if ( ! ipiobuf_can_pull(iobufP, sizeof(ashort))) return 0;

   memcpy(&ashort,iobufP->data_array+iobufP->pull_index,sizeof(ashort));

   iobufP->pull_index += sizeof(ashort);

   return ashort;
   }

/**
 * pull a uint32 from the iobuf
 * NOTE that uint16 (and possibly longer integers MUST be even aligned)
 * @return the uint32 value
 */
uint32 ipiobuf_pull_uint32 (struct Ipiobuf * iobufP)
   {
   uint32 aval;

   ipiobuf_pull_even(iobufP);  // if the push has an even, the pull has it too

   // if there is not enough to pull
   if ( ! ipiobuf_can_pull(iobufP, sizeof(aval))) return 0;

   memcpy(&aval,iobufP->data_array+iobufP->pull_index,sizeof(aval));

   iobufP->pull_index += sizeof(aval);

   return aval;
   }


/**
 * push a byte to the iobuf
 * @return the number of bytes written or zero
 */
int ipiobuf_push_uint8 (struct Ipiobuf * iobufP, uint8 abyte)
   {
   if ( ! ipiobuf_can_push(iobufP, 1)) return 0;

   iobufP->data_array[iobufP->push_index++] = abyte;

   return 1;
   }

/**
 * push a UTF8 string to wire
 * remember that a UTF8 cannot have a zero in it, so, it is safe to send it as a "unix" string
 * ALSO, if the string does not fit, a PARTIAL string will be sent, it WILL be zero terminated
 * @return the number of bytes written or zero
 */
int ipiobuf_push_utf8s (struct Ipiobuf * const iobufP, void *utf8P)
   {
   if ( ! utf8P ) return 0;

   int s_len=strlen(utf8P);

   int can_push=ipiobuf_get_push_max(iobufP);

   // cannot push anything, not even a byte
   if ( can_push < 1 ) return 0;

   if ( s_len >= can_push )
      {
      // NOTE that it is a >= meaning that I reserve ONE byte at the end
      // I know can_push >= 1 AND I need to reserve a zero byte at the end
      s_len=can_push-1;
      }

   uint8 *data_array = iobufP->data_array;
   uint8 *suP = utf8P;

   while ( s_len-- > 0 )
      data_array[iobufP->push_index++] = *suP++;

   data_array[iobufP->push_index++] = 0;

   return 1;
   }


/**
 * NOTE that it will return the actual structure of a linkid, NOT the chunk header part
 * so, this is valid even when the chunk header is 16 bits
 * @return a linkid if available, or null
 */
struct IppLinkid *ipiobuf_get_linkid(struct Ipiobuf *ibufP )
   {
   if ( ! ibufP ) return NULL;

   if ( ! ibufP->linkid_index ) return NULL;

   void *risulP = ibufP->data_array + ibufP->linkid_index;

   return risulP;
   }


/**
 * GET, meaning that it gets the value but does NOT advance the pull index
 * @return the uint8 value
 */
uint8 ipiobuf_get_uint8 (struct Ipiobuf * const iobufP)
   {
   return iobufP->data_array[iobufP->pull_index];
   }

/**
 * GET, meaning that it gets the value at the given index
 * Check with ALL the places where it is used that iobufP is non null
 * @return the uint8 value
 */
uint8 ipiobuf_get_uint8_at (struct Ipiobuf * const iobufP, uint16 at_index)
   {
   return iobufP->data_array[at_index];
   }

uint8 *ipiobuf_get_uint8P_at (struct Ipiobuf * const iobufP, uint16 at_index)
   {
   return iobufP->data_array+at_index;
   }

/**
 * pull a uint8 from the iobuf
 * @return the uint8 value
 */
uint8 ipiobuf_pull_uint8 (struct Ipiobuf * iobufP)
   {
   // if there is nothing to pull
   if ( ! ipiobuf_can_pull_something(iobufP)) return 0;

   return iobufP->data_array[iobufP->pull_index++];
   }

/**
 * Attempt to pull the given amount of bytes
 * Just advance the pull_index if it is possible
 */
int ipiobuf_pull_skip (struct Ipiobuf * iobufP, uint16 n_bytes)
   {
   if ( ! ipiobuf_can_pull(iobufP, n_bytes)) return 0;

   iobufP->pull_index += n_bytes;

   return n_bytes;
   }


/**
 * Push a number of bytes reservation
 * note that this works on any part of the buffer, not just at the beginning
 * @return the pointer to the beginning of the reserved area or NULL if no space
 */
void *ipiobuf_push_reserve (struct Ipiobuf * const iobufP, uint16 n_bytes)
   {
   if ( ! iobufP ) return NULL;

   if ( ! ipiobuf_can_push(iobufP, n_bytes)) return NULL;

   // I need to return the beginning of the allocated buffer
   void *risulP = iobufP->data_array+iobufP->push_index;

   // this is how you reserve space
   iobufP->push_index += n_bytes;

   return risulP;
   }

/**
 * push a chunk8 header
 * @return the pointer to the length (that needs to be adjusted)
 */
uint8 *ipiobuf_push_chunk8_header (struct Ipiobuf * iobufP, enum IppChunkid chunk_type)
   {
   if ( ! ipiobuf_can_push(iobufP, sizeof(struct IppChunkh8))) return NULL;

   // chunk8 MUST have the LSB at 0 I better catch it sooner than later !
   if ( chunk_type & 0x01 )
      return NULL;

   iobufP->data_array[iobufP->push_index++] = chunk_type;

   void *lenP = ipiobuf_get_pushP(iobufP);

   iobufP->data_array[iobufP->push_index++] = 0;  // this is the length, it will be adjusted at the end

   return lenP;
   }

/**
 * push a chunk16 header
 * @return the pointer to the length (that needs to be adjusted)
 */
uint16 *ipiobuf_push_chunk16_header (struct Ipiobuf * iobufP, enum IppChunkid chunk_type )
   {
   if ( ! ipiobuf_can_push(iobufP, sizeof(struct IppChunkh16))) return NULL;

   // chunk16 MUST have LSB bit at 1 I better catch it sooner than later !
   if ( ! (chunk_type & 0x01) )
      return NULL;

   iobufP->data_array[iobufP->push_index++] = chunk_type;
   iobufP->data_array[iobufP->push_index++] = 0;

   void *lenP = ipiobuf_get_pushP(iobufP);

   iobufP->data_array[iobufP->push_index++] = 0;
   iobufP->data_array[iobufP->push_index++] = 0;

   return lenP;
   }

/**
 * this assumes that all the chunktype are 16 bits
 * Used for the dialog between the router and the appx
 */
uint16 *ipiobuf_push_chunkx_header (struct Ipiobuf * iobufP, enum IppHRchunkid chunk_type )
   {
   if ( ! ipiobuf_can_push(iobufP, sizeof(struct IppChunkh16))) return NULL;

   iobufP->data_array[iobufP->push_index++] = chunk_type;
   iobufP->data_array[iobufP->push_index++] = 0;

   void *lenP = ipiobuf_get_pushP(iobufP);

   iobufP->data_array[iobufP->push_index++] = 0;
   iobufP->data_array[iobufP->push_index++] = 0;

   return lenP;
   }


/**
 * push an array of bytes into the buffer
 * this will work even with length zero, meaning that nothing will be pushed
 * if the available space is LESS than src_len only PART of data will be pushed
 * @return the new point where you can write
 */
void *ipiobuf_push_array(struct Ipiobuf * const iobufP, void * const srcP, uint16 src_len )
   {
   int can_push=ipiobuf_get_push_max(iobufP);

   if ( src_len > can_push )
      src_len=can_push;

   uint8 *data_array = iobufP->data_array;
   uint8 *suP = srcP;

   while ( src_len-- > 0 )
      data_array[iobufP->push_index++] = *suP++;

   return data_array+iobufP->push_index;
   }

/**
 * push an array of bytes into the buffer
 * this will work even with length zero, meaning that nothing will be pushed
 * this is the one to be used when receiving from net !!!
 * @return the new point where you can write
 */
void *ipiobuf_rxnet_push_array(struct Ipiobuf * const iobufP, void * const srcP, uint16 src_len )
   {
   int can_push=ipiobuf_get_packet_len(iobufP) - iobufP->push_index;

   if ( src_len > can_push )
      src_len=can_push;

   uint8 *data_array = iobufP->data_array;
   uint8 *suP = srcP;

   while ( src_len-- > 0 )
      data_array[iobufP->push_index++] = *suP++;

   return data_array+iobufP->push_index;
   }


/**
 * pull a UTF8 string from the buffer, stops at the first zero or when the no more
 * this works ok even with to_len equal zero
 * NOTE that there MUST be a zero at the end of string !
 * @return the number of bytes copied or zero
 */
int ipiobuf_pull_utf8s(struct Ipiobuf * const iobufP, void *tosP, uint16 const to_len )
   {
   if ( ! tosP ) return 0;

   // I do not have space, not even for the simple 0 at the end
   if ( to_len < 1 ) return 0;

   // this is what ideally I wish
   uint16 w_len=to_len-1;

   uint16 pullmax=ipiobuf_get_pull_max(iobufP);

   // in ANY case, do NOT get more that what is in the buffer
   if ( pullmax < w_len ) w_len=pullmax;

   int letti=0;

   uint8 *data_array = iobufP->data_array;
   uint8 *toP = tosP;

   uint8 achar=0;   // I need it out since I need to test if it is zero

   while ( w_len-- > 0 )
      {
      achar=data_array[iobufP->pull_index++];
      *toP++ = achar;  // ALSO need to copy the zero at the end
      // IF I have found the zero termination, end of all work
      if ( achar == 0 ) break;
      // letti does not include the zero termination
      letti++;
      }

   // if we exited because of a zero
   if ( achar == 0 ) return letti;

   // this means that for some reason I did NOT read the whole string in the buffer aand I MUST read all of it
   *toP++ = 0;  // make sure the result is zero terminated

   // now, try to get to the end of string in the buffer
   pullmax=ipiobuf_get_pull_max(iobufP);

   while ( pullmax-- > 0 )
      // IF I have found the zero termination, end of all work
      if ( data_array[iobufP->pull_index++] == 0 )
         break;

   return letti;
   }


/**
 * this works ok even with to_len equal zero
 * it ALSO works ok if to_len is biffer than pull_max since it copies only what is available
 * @return the number of bytes copied or zero
 */
int ipiobuf_pull_array(struct Ipiobuf * const iobufP, void *tovP, uint16 to_len )
   {
   int  pull_max=ipiobuf_get_pull_max(iobufP);

   if ( to_len > pull_max )
      to_len=pull_max;

   int retcode = to_len;

   uint8 *data_array = iobufP->data_array;
   uint8 *toP = tovP;

   while ( to_len-- > 0 )
      *toP++ = data_array[iobufP->pull_index++];

   return retcode;
   }


/**
 * this is easier, since the total length is actually what I should write
 * @return the number of bytes written
 */
int ipiobuf_write_to (struct Ipiobuf *  iobufP, int unix_fd)
   {
   int txmax = ipiobuf_get_pull_max(iobufP);

   if ( txmax <= 0 ) return 0;

   int scritti = write(unix_fd, ipiobuf_get_pullP(iobufP), txmax);

   if ( scritti <= 0 ) return 0;

   iobufP->pull_index += scritti;

   return scritti;
   }


/**
 * Pulls a chunk header and returns a pointer to it
 * so, you can read the rest of data keeping an eye to the header
 * @return a pointer to the chunk or NULL if not enough  bytes
 */
struct IppChunkh8 *ipiobuf_pull_chunkh8(struct Ipiobuf * const bufP)
   {
   if ( ! bufP ) return NULL;

   // two bytes of packet len and two bytes of the header
   if ( ipiobuf_get_pull_max(bufP) < sizeof(struct IppChunkh8) ) return NULL;

   // the chunk header is always at the current index
   void *risulP = bufP->data_array+bufP->pull_index;

   // the first chunk is always a chunk8
   bufP->pull_index += sizeof(struct IppChunkh8);

   return risulP;
   }


/**
 * Pulls a chunk 16 header and returns a pointer to it
 * so, you can read the rest of data keeping an eye to the header
 * do NOT check if the chunkid is correct since this one is used even for the data from app to router !!
 * @return a pointer to the chunk or NULL if not enough  bytes
 */
struct IppChunkh16 *ipiobuf_pull_chunkh16(struct Ipiobuf * const bufP)
   {
   if ( ! bufP ) return NULL;

   // two bytes of packet len and two bytes of the header
   if ( ipiobuf_get_pull_max(bufP) < sizeof(struct IppChunkh16) ) return NULL;

   // the chunk header is always at the current index
   void *risulP = bufP->data_array+bufP->pull_index;

   // then advance to next content
   bufP->pull_index += sizeof(struct IppChunkh16);

   return risulP;
   }

/**
 * if you scan over the packet and want to pull again
 */
void ipiobuf_pull_reset(struct Ipiobuf * const bufP)
   {
   bufP->pull_index=0;
   }

/**
 * Returns a pointer to the chunk header AND it advances the pull_index after the header
 * so, you can read the rest of data keeping an eye to the header
 * This is for the eight bit chunks !
 * @return a pointer to the chunk or NULL if not enough  bytes
 */
struct IppChunkh8 *ipiobuf_pull_next_chunk8(struct Ipiobuf * bufP)
   {
   if ( ! bufP ) return NULL;

   // two bytes of the chunk header
   if ( ipiobuf_get_pull_max(bufP) < 2 ) return NULL;

   // I am right at the beginning of the chunk
   void *risulP = bufP->data_array+bufP->pull_index;

   // we are dealing with chunk8
   bufP->pull_index += 2;

   return risulP;
   }


/**
 * Retrieve the next chunk, NOTE that chunk may be 8 or 16 bits !
 * The first call you pass prevP->n_chunkP=NULL
 * Note that it returns where you should do the looup !!
 */
struct ChunkIterator *ipiobuf_next_chunk(struct Ipiobuf * const bufP, struct ChunkIterator * const iterP)
   {
   if ( ! bufP || ! iterP ) return NULL;

   if ( ! iterP->n_chunkP )
      {
      // this is the first call
      memzero_ptr(iterP);
      iterP->packetlen=ipiobuf_pull_packet_len(bufP);
      }

   if ( ! ipiobuf_can_pull(bufP, sizeof(struct IppChunkh8)) )
      {
      // this is the end of the scan, cannot pull anything more
      iterP->chunklen=0;
      iterP->chunkType=0;      // this signal the end of loop
      iterP->chunkDataP=NULL;
      iterP->n_chunkP=NULL;
      return iterP;
      }

   // note that it is a GET, not a pull !!
   uint8 chunkid = ipiobuf_get_uint8(bufP);

   if ( chunkid == ICHUNK_ONEBYTE )
      {
      // this chunkid is used to align even the chunks
      // skip the one byte chunk
      ipiobuf_pull_uint8(bufP);
      // and get the real chunkid
      chunkid = ipiobuf_get_uint8(bufP);
      }

   if ( chunkid & 0x01 )
      {
      struct IppChunkh16 *chuP = ipiobuf_pull_chunkh16(bufP);
      iterP->chunkhLen = sizeof(struct IppChunkh16);
      iterP->chunkType = chuP->chunk_type;
      iterP->chunklen  = chuP->chunk_len;
      }
   else
      {
      struct IppChunkh8 *chuP = ipiobuf_pull_chunkh8(bufP);
      iterP->chunkhLen = sizeof(struct IppChunkh8);
      iterP->chunkType = chuP->chunk_type;
      iterP->chunklen  = chuP->chunk_len;
      }

   iterP->chunkDataP = ipiobuf_get_pullP(bufP);

   ipiobuf_pull_skip(bufP,iterP->chunklen-iterP->chunkhLen);

   iterP->n_chunkP = ipiobuf_get_pullP(bufP);

   return iterP;
   }

/**
 * the first hunk id is always at the same place
 * Since the FIRST chunk is always an eight bit len
 */
uint8 ipiobuf_get_first_chunkid (struct Ipiobuf *ibufP )
   {
   if ( ! ibufP ) return 0;

   return ibufP->data_array[2];
   }


