/**
 * Copyright Damiano Bolla, Cologna Veneta, Italy, All right reserved
 * Eclipse does not have any more copyright manager, so I need to edit by hand
 * GNU LESSER GENERAL PUBLIC LICENSE Version 3, 29 June 2007
 */

#ifndef IPIOBUF_H
#define IPIOBUF_H

#include <hal.h>
#include <ipp_sock.h>

/**
 * the idea is that an iobuf can be partially read and written
 * a iobuf has a storage area and a write index + read index
 * a iobuf should be malloc, so I can send it around the system
 *
 * I can PUSH data into the buffer
 * I can PULL data out of the buffer
 */

#define IPIOBUF_PKT_MAX       1600     // the idea is that a frame should fit into ethernet, no jumbo !
#define IPIOBUF_MAGIC_H       0x33     // A magic code to identify if a void * is actually an ipiobuf
#define IPIOBUF_MAGIC_NAKED   0x44     // A magic code to identify an ipiobuf of naked data
#define IPIOBUF_MULTIPLEN     1024     // size must be a multiple of this

/**
 * NOTE on packet_len, when constructing a packet it can only be determined at the end
 * on reception, after the first two bytes, it is known
 * a value of zero, means that it is not known, yet
 * NOTE that this is shared between router code and "socket" implementation
 * the user application must NOT consider this an API
 */

#define IPIOBUF_HEADER_LEN    16    // sizeof(struct Ipiobuf) MUST return this number, compiler sanity check


union IpiobufFlags
   {
   struct
      {
      uint8 clear_tx_queue:1;    // If set this ipiobuf is a discontinuity on the TX queue
      uint8 reserved:7;          // MSB bits for LE CPU
      } bits;

   uint8 val;
   } STRUCT_PACKED;

struct Ipiobuf
   {
   uint8  object_type;      // safety to check if this is somewhat valid
   uint8  object_refcount;  // once refcount goes to zero, it can be deallocated
   uint8  tx_try_count;     // used to decide actions on excessive retransmit
   union  IpiobufFlags ipio_flags;
   uint16 ipface_idx;       // is directed to the given ipface
   uint16 soface_idx;       // buffer to be given to this soface index
   uint16 pull_index;       // start reading from here when reading the buffer, PULL
   uint16 push_index;       // legal write location when filling the buffer, PUSH
   uint16 data_array_len;   // the MAX length of data array in the ipiobufT, allocated on creation
   uint16 linkid_index;     // IF there is a linkid in the buffer, this is the beginning, 0 means NO linkid present
   uint8  data_array[0];    // the size of this should be normally zero, but by malloc you will extend
   } STRUCT_PACKED;

struct ChunkIterator
   {
   void *n_chunkP;     // pointer to the beginning of next chunk. NULL on first call
   void *chunkDataP;   // pointer to the actual data
   uint16 packetlen;   // as written at the beginning of packet
   uint16 chunklen;
   uint8  chunkType;
   uint8  chunkhLen;    // how long is the chunk header len ?4 or 2 bytes
   uint8  i_flags;      // like have 16bits datalen
   };

struct ChunkIterator *ipiobuf_next_chunk(struct Ipiobuf * const bufP, struct ChunkIterator * const prevP);


int ipiobuf_init(void);
void ipiobuf_print_stats(void);

struct Ipiobuf * ipiobuf_malloc(uint8 objtype, uint16 size);
struct Ipiobuf * ipiobuf_release ( struct Ipiobuf * iobufP );

int ipiobuf_rxnet_from (struct Ipiobuf *  iobufP, int unix_fd);
int ipiobuf_rxnet_max (struct Ipiobuf *  iobufP, int unix_fd, int w_max );
int ipiobuf_rxnet_get_push_max(struct Ipiobuf *const iobufP);

void ipiobuf_pull_reset(struct Ipiobuf * const bufP);
int ipiobuf_get_pull_max(struct Ipiobuf *  iobufP);
int ipiobuf_can_pull_something (struct Ipiobuf * iobufP );
int ipiobuf_can_pull(struct Ipiobuf *  iobufP, uint16 nbytes );
int ipiobuf_write_to (struct Ipiobuf * iobufP, int unix_fd);

void *ipiobuf_get_pushP(struct Ipiobuf * iobufP);
void *ipiobuf_get_pullP(struct Ipiobuf *  iobufP);
struct IppLinkid *ipiobuf_get_linkid(struct Ipiobuf *ibufP );

int ipiobuf_can_push(struct Ipiobuf * iobufP, uint16 n_bytes);
int ipiobuf_can_push_something(struct Ipiobuf * iobufP);

void *ipiobuf_push_reserve (struct Ipiobuf * iobufP, uint16 n_bytes);

enum IppChunkid;  // forward declare the enum
uint8  *ipiobuf_push_chunk8_header  (struct Ipiobuf * iobufP, enum IppChunkid chunk_type);
uint16 *ipiobuf_push_chunk16_header (struct Ipiobuf * iobufP, enum IppChunkid chunk_type );
uint16 *ipiobuf_push_chunkx_header (struct Ipiobuf * iobufP, enum IppHRchunkid chunk_type );

void *ipiobuf_rxnet_push_array(struct Ipiobuf * const iobufP, void * const srcP, uint16 src_len );
void *ipiobuf_push_array  (struct Ipiobuf * iobufP, void *srcP, uint16 src_len );

int  ipiobuf_push_uint8  (struct Ipiobuf * iobufP, uint8 abyte);
int  ipiobuf_push_uint16 (struct Ipiobuf * iobufP, uint16 ashort);
int  ipiobuf_push_uint32 (struct Ipiobuf * iobufP, uint32 aval);
int  ipiobuf_push_utf8s  (struct Ipiobuf * const iobufP, void *utf8P);
int  ipiobuf_push_even   (struct Ipiobuf * iobufP );

uint16 ipiobuf_pull_packet_len(struct Ipiobuf * bufP);
uint8  ipiobuf_pull_uint8  (struct Ipiobuf * iobufP);
uint16 ipiobuf_pull_uint16 (struct Ipiobuf * iobufP);
uint32 ipiobuf_pull_uint32 (struct Ipiobuf * iobufP);

uint8 ipiobuf_get_uint8 (struct Ipiobuf * iobufP);
uint8 ipiobuf_get_uint8_at (struct Ipiobuf * const iobufP, uint16 at_index);
uint8 *ipiobuf_get_uint8P_at (struct Ipiobuf * const iobufP, uint16 at_index);
uint8 ipiobuf_get_first_chunkid (struct Ipiobuf *ibufP );

int ipiobuf_pull_array  (struct Ipiobuf * iobufP, void *tovP, uint16 to_len );
int ipiobuf_pull_even   (struct Ipiobuf * iobufP);
int ipiobuf_pull_skip   (struct Ipiobuf * iobufP, uint16 n_bytes);
int ipiobuf_pull_utf8s  (struct Ipiobuf * const iobufP, void *tosP, uint16 const to_len );

uint8  ipiobuf_set_chulen8 (struct Ipiobuf *iobufP, uint8  *clenP );
uint16 ipiobuf_set_chulen16(struct Ipiobuf *iobufP, uint16 *clenP );

uint16 ipiobuf_set_pktlen_new(struct Ipiobuf * iobufP );
uint16 ipiobuf_rxnet_push_packet_len(struct Ipiobuf * const ibufP, uint16 rx_pktlen );
uint16 ipiobuf_set_packet_len(struct Ipiobuf * const ibufP, uint16 pktlen );

uint16 ipiobuf_get_packet_len(struct Ipiobuf * bufP);
int ipiobuf_rxnet_is_received(struct Ipiobuf * const iobufP);

struct IppChunkh8 *ipiobuf_pull_chunkh8(struct Ipiobuf * bufP);
struct IppChunkh8 *ipiobuf_pull_next_chunk8(struct Ipiobuf * bufP);

struct IppChunkh16 *ipiobuf_pull_chunkh16(struct Ipiobuf * const bufP);




#endif /* IOBUF_C_ */
