/**
 * Copyright Damiano Bolla, Cologna Veneta, Italy, All right reserved
 * Eclipse does not have any more copyright manager, so I need to edit by hand
 * GNU LESSER GENERAL PUBLIC LICENSE Version 3, 29 June 2007
 */

#include <ipp_sock.h>

#include <strtoksep.h>
#include <stringd.h>
#include <ipiobuf.h>
#include <ipps_utils.h>

#include <fcntl.h>
#include <sys/poll.h>

#define IPPSOC_DEBUG_IPV4CONN    0x01
#define IPPSOC_DEBUG_ACCEPT      0x02
#define IPPSOC_DEBUG_TCPCONNECT  0x04
#define IPPSOC_DEBUG_THREADLIFE  0x08

/**
 * the reason to pass the pointer to where data will be saved is to allow for user handling of data allocation
 * this way it can be all static or dynamic, as anyone please
 * It also solves issues of memory allocation, meaning, that issues are left to the application
 */
static int prisoc_set_error(struct IPP_soc *const socP, int const error, char *message )
   {
   socP->soc_errcode=error;
   strncpyda(socP->soc_errmsg, message,sizeof(socP->soc_errmsg));
   return error;
   }

static int prisoc_isDebugMask (struct IPP_soc *const socP, uint32 w_mask )
   {
   return (socP->debug_mask & w_mask) == w_mask;
   }

static int prisoc_valid_proto(enum IppChunkid want )
   {
   switch ( want )
      {
      case ICHUNK_UDP:
      case ICHUNK_TCP:
      case ICHUNK_ICMP:
      case ICHUNK_GENERIC:
         return TRUE;

      default:
         return FALSE;
      }
   }


/**
 * MUST init the socket for use !
 * NOTE: do NOT init ipiobuf since otherwise it will be init many times
 * you HAVE to call open to actually start the work
 * @param rtr_port the ipv4 port to connect to router, if 0 the default value of 2024 will be used
 */
int ippsoc_init(struct IPP_soc *const socP, enum IppChunkid w_proto, char *rtr_ipv4, uint16 rtr_port, uint32 debug_mask)
   {
   if ( ! socP )
      {
      fprintf(stderr,"ippsoc_init: ERROR sockP is NULL\n");
      return IPPS_ERR_NULLP;
      }

   memzero_ptr(socP);

   if ( ! prisoc_valid_proto(w_proto))
      return prisoc_set_error(socP, IPPS_ERR_BADPROTO, "ippsoc_init: unsupported protocol");

   if ( ! rtr_port )
      rtr_port=2024;

   socP->soc_proto=w_proto;
   socP->debug_fd=stderr;
   socP->debug_mask=debug_mask;
   socP->qw_timeout_s=20;    // you can change it between init and open

   if ( sizeof(struct IppChunkh16) != IPPS_hrs_header_LEN )
      return prisoc_set_error(socP, IPPS_ERR_BSIZEOF, "INVALID sizeof IPPS_hs_header");

   if ( unix_inetv4_aton(&socP->rtrc_addr,rtr_ipv4,rtr_port) )
      return prisoc_set_error(socP, IPPS_ERR_BADV4, "cannot convert ipv4 address");

   return DONE;
   }

/**
 * Init the given socket with info from the other socket
 */
static int prisoc_init(struct IPP_soc *const socP, struct IPP_soc *const fromP)
   {
   memzero_ptr(socP);

   socP->soc_proto=fromP->soc_proto;
   socP->debug_fd=fromP->debug_fd;
   socP->debug_mask=fromP->debug_mask;
   socP->rtrc_addr=fromP->rtrc_addr;     // this copies the whole address, including the port
   socP->qw_timeout_s=fromP->qw_timeout_s;

   return DONE;
   }

/**
 * Just connect to the router, TCP connection
 * If connection is already done, just return DONE
 * This could set timeout, probably a good thing...
 * @return 0 on success or some error code
 */
static int prisoc_connect_router(struct IPP_soc *const sockP)
   {
   if ( prisoc_isDebugMask(sockP, IPPSOC_DEBUG_IPV4CONN))
      fprintf(sockP->debug_fd,"prisoc_connect_router: CALL \n");

   if ( sockP->router_fd ) return DONE;

   // I am ALWAYS on stream, UDP is over TCP....
   sockP->router_fd = socket(AF_INET, SOCK_STREAM, 0);

   if (sockP->router_fd < 0)
      return prisoc_set_error(sockP,IPPS_ERR_NO4SO,"cannot create socket");

   // when a connection is naked, it is OK to leave it open to children, exec
   //fcntl_set_close_on_exec ( sockP->router_fd );

   if ( connect(sockP->router_fd,(struct sockaddr *)&sockP->rtrc_addr,sizeof(sockP->rtrc_addr)) )
      {
      close (sockP->router_fd);
      sockP->router_fd=0;
      return prisoc_set_error(sockP,IPPS_ERR_NOCONN,"router NOT connected");
      }

   return DONE;
   }


static int prisoc_read_router_reply(struct IPP_soc *const socP );


static void *prisoc_io_fun(void *argP)
   {
   struct IPP_soc *const socP = argP;

   if ( prisoc_isDebugMask(socP, IPPSOC_DEBUG_THREADLIFE))
      fprintf(socP->debug_fd,"prisoc_io_fun: START\n");

   while ( ! ( socP->router_eof || socP->soc_naked ) )
      prisoc_read_router_reply(socP);

   if ( prisoc_isDebugMask(socP, IPPSOC_DEBUG_THREADLIFE))
      fprintf(socP->debug_fd,"prisoc_io_fun: END idx=%d\n",socP->soface_idx);

   return NULL;
   }


int ippsoc_open(struct IPP_soc *const socP, IppsocRxparseFun funP)
   {
   if ( ! socP )
      return IPPS_ERR_NULLP;

   if ( socP->read_from_router_thread )
      return prisoc_set_error(socP, IPPS_ERR_BADSTATE, "ippsoc_open: socket is already opened");

   if ( prisoc_connect_router(socP) )
      return prisoc_set_error(socP, IPPS_ERR_BADSTATE, "ippsoc_open: cannot connect to router");

   if ( pthread_mutex_init(&socP->qw_mutex, NULL) )
      return prisoc_set_error(socP, IPPS_ERR_BADSTATE, "ippsoc_open: pthread_mutex_init FAILED");

   if (pthread_cond_init(&socP->qw_cond, NULL))
      return prisoc_set_error(socP, IPPS_ERR_BADSTATE, "ippsoc_open: pthread_cond_init FAILED");

   // this thread should shuffle IO and parse packets...
   if ( pthread_create(&socP->read_from_router_thread, NULL, prisoc_io_fun, socP) )
      return prisoc_set_error(socP, IPPS_ERR_BADSTATE, "ippsoc_open: pthread_create FAILED");

   return ippsoc_set_uparsefun(socP, funP);
   }


int ippsoc_set_uparsefun(struct IPP_soc *const socP, IppsocRxparseFun funP)
   {
   if ( ! socP ) return IPPS_ERR_NULLP;

   socP->uparseFun=funP;

   return DONE;
   }

int ippsoc_print_stats(struct IPP_soc *const socP)
   {
   ipiobuf_print_stats();

   if ( ! socP ) return IPPS_ERR_NULLP;

   CONSOLE_PRINTF("soc_proto... = %d\n",socP->soc_proto);
   CONSOLE_PRINTF("sock fd..... = %d\n",socP->router_fd);
   CONSOLE_PRINTF("soface_idx.. = %d\n",socP->soface_idx);
   CONSOLE_PRINTF("sock_errcode = %d\n",socP->soc_errcode);
   CONSOLE_PRINTF("sock_errmsg. = %s\n",socP->soc_errmsg);
   CONSOLE_PRINTF("naked....... = %d\n",socP->soc_naked);

   if ( socP->soc_proto != ICHUNK_TCP ) return DONE;

   struct IppTcpState *tsP = &socP->tcp_state;

   CONSOLE_PRINTF("TCP to port..... = %d\n",tsP->dest_port);
   CONSOLE_PRINTF("  last_rxed_seq. = %d\n",tsP->last_rxed_seq);
   CONSOLE_PRINTF("  this_pk_seq... = %d\n",tsP->this_pk_seq);

   return DONE;
   }


int ippsoc_is_ipv4connect(struct IPP_soc *const sockP)
   {
   return sockP->router_fd != 0;
   }

/**
 * Public interface
 */
int ippsoc_close (struct IPP_soc *socP )
   {
   if ( ! socP )
      return IPPS_ERR_NULLP;

   if ( prisoc_isDebugMask(socP, IPPSOC_DEBUG_IPV4CONN))
      fprintf(socP->debug_fd,"ippsoc_close: CALL port %d\n",socP->soc_to_port);

   if ( socP->router_fd )
      close (socP->router_fd);

   socP->router_fd=0;

   socP->router_eof=TRUE;

   // TODO: I should also deallocate the queue

   return DONE;
   }

static int prisoc_tx(struct IPP_soc *const sockP)
   {
   if ( ! sockP->router_fd )
      return 0;

   int txmax = ipiobuf_get_pull_max(sockP->tx_bufP);

   if ( txmax < 1 )
      {
      // it may happen that an empty packet is queued ... should not happen ?
      sockP->tx_bufP = ipiobuf_release(sockP->tx_bufP);
      // BUT it is NOT an EOF
      return 0;
      }

   int scritti=ipiobuf_write_to(sockP->tx_bufP, sockP->router_fd);

   if ( scritti <= 0 )
      {
      // this also deallocate the buffers
      ippsoc_close(sockP);
      return 0;
      }

   txmax = ipiobuf_get_pull_max(sockP->tx_bufP);

   if ( txmax < 1 )
      {
      // for small packets, the write will do all is needed
      sockP->tx_bufP = ipiobuf_release(sockP->tx_bufP);
      }

   return scritti;
   }

/**
 * This can ONLY return DONE or an error code, for connection terminated
 */
static int prisoc_tx_all_C(struct IPP_soc *const sockP)
   {
   while (ipiobuf_get_pull_max(sockP->tx_bufP) > 0 )
      if ( prisoc_tx(sockP) <= 0 )
         {
         ippsoc_close(sockP);
         return IPPS_ERR_WRITE4FAIL;
         }

   return IPPS_DONE;
   }

/**
 * write to router the given request, all of it
 * we assume to be in LAN and that we can wait for full write
 * NOTE that the buffer WILL be released, regardless of error code
 * @return DONE or an error code, negative
 */
static int prisoc_tx_all_AA(struct IPP_soc *const sockP, struct Ipiobuf *const bufP, char *done_msg)
   {
   if ( sockP->tx_bufP )
      return prisoc_set_error(sockP,IPPS_ERR_TXOVERRUN,"TX twice(same time) on socket");

   sockP->tx_bufP = bufP;

   int errcode=prisoc_tx_all_C(sockP);

   sockP->tx_bufP = ipiobuf_release(sockP->tx_bufP);

   return prisoc_set_error(sockP,errcode,done_msg);
   }

/**
 * attempt to read the first two bytes of the packet len
 */
static int prisoc_rx_pklen(struct IPP_soc *const sockP)
   {
   uint16 pkt_len;

   if ( sockP->rx_bufP )
      {
      // the policy is that the buffer is released here...
      sockP->rx_bufP = ipiobuf_release(sockP->rx_bufP);
      }

   // I just need to read the whole packet length
   int letti = read(sockP->router_fd, &pkt_len, sizeof(pkt_len));

   if ( letti < 2 )
      {
      ippsoc_close(sockP);
      return prisoc_set_error(sockP,IPPS_ERR_SOCKCLO,"prisoc_rx_pklen: A socket closed");
      }

   // I can allocate a buffer bigger, since the test is done against the actual packet len
   sockP->rx_bufP = ipiobuf_malloc(IPIOBUF_MAGIC_H,pkt_len+100);

   if ( ! sockP->rx_bufP )
      {
      // if the iobuffer is still NULL then abort
      ippsoc_close(sockP);
      return prisoc_set_error(sockP,IPPS_ERR_NOMALLOC,"prisoc_rx_pklen: cannot malloc");
      }

   // MUST push the incoming packet size
   ipiobuf_rxnet_push_packet_len(sockP->rx_bufP, pkt_len);

   return DONE;
   }

/**
 * Someone else MUST set the socP->qw_for with the right value !!
 * BEFORE calling the signal
 */
static int prisoc_qwait_timed ( struct IPP_soc *const socP, enum IppHRchunkid want_event, uint16 wait_s )
   {
   struct timespec w_time;

   // we already know we have data, no need to wait
   if ( socP->qw_have_event == want_event ) return TRUE;

   // Need to get a lock on something in any case
   if ( pthread_mutex_lock(&socP->qw_mutex) ) return FALSE;

   clock_gettime(CLOCK_REALTIME, &w_time);

   w_time.tv_sec += wait_s;

   while ( socP->qw_have_event != want_event )
      {
      // apparently the function may return 0 even if no data is present
      int rcode = pthread_cond_timedwait(&socP->qw_cond, &socP->qw_mutex, &w_time);

      // but if there is a timeout, it is really a timeout
      if ( rcode == ETIMEDOUT ) break;
      }

   pthread_mutex_unlock(&socP->qw_mutex);

   return socP->qw_have_event == want_event;
   }

/**
 * signal to whoever is waiting that a given event has been received
 */
static int prisoc_qwait_signal ( struct IPP_soc *const socP, enum IppHRchunkid have_event )
   {
   socP->qw_have_event=have_event;

   return pthread_cond_broadcast(&socP->qw_cond);
   }

/**
 * normally return zero, if all is fine
 */
static int prisoc_parse_make_naked_rh(struct IPP_soc *const socP, int rescode)
   {
   if ( rescode == DONE )
      {
      // MUST say that this socket is naked, now
      socP->soc_naked=TRUE;
      // then call the signal function
      return prisoc_qwait_signal(socP, IPPHRchu_make_naked_rh);
      }

   if ( prisoc_isDebugMask(socP, IPPSOC_DEBUG_ACCEPT))
      fprintf(socP->debug_fd,"prisoc_parse_make_naked_rh: idx=%d rescode=%d msg=%s",socP->soface_idx, rescode, socP->rtr_errmsg);

   return rescode;
   }

/**
 * remember that the user can also have a look at the packets
 */
static void prisoc_message_rh(struct IPP_soc *const socP, struct Ipiobuf *const bufP)
   {
   struct IPPS_hrp_reply *resP = ipiobuf_get_pullP(bufP);
   socP->soface_idx=resP->soface_idx;
   socP->rtr_errcode=resP->hrp_rescode;
   strncpyda(socP->rtr_errmsg,resP->hrp_message,sizeof(socP->rtr_errmsg));

   if ( resP->o_request == IPPHRchu_make_naked_rh )
      prisoc_parse_make_naked_rh(socP,resP->hrp_rescode );

   }

/**
 * From CLIENT -> SERVER
 * this is the client that is sending a request to connect
 * server will use this info with the accept function
 * At the moment I only store one request
 */
static void prisoc_parse_tcp_syn_cs(struct IPP_soc *const sockP, struct Ipiobuf *const bufP)
   {
   struct IPPS_tcp_accept_info *rxP = ipiobuf_get_pullP(bufP);
   struct IppTcpState *const tcpshP = &rxP->tcp_state;

   // set aside this info for later use
   sockP->acceptw_array[0] = *rxP;

   if ( prisoc_isDebugMask(sockP, IPPSOC_DEBUG_ACCEPT))
      fprintf(sockP->debug_fd,"prisoc_parse_tcp_syn_cs: soface=%d \n",rxP->soface_idx);

   tcpshP->tcp_sostate=SOTCPC_SYN_RECV;

   // if anybody is waiting for this event, signal it
   prisoc_qwait_signal(sockP, IPPHRchu_tcp_rh_syn_cs);
   }

/**
 * I need this since there are events that happens automatically, especially for TCP handling
 * But it could also be for OOB data or ICMP
 */
static void prisoc_parse_sock_reply(struct IPP_soc *const socP)
   {
   struct Ipiobuf *const bufP = socP->rx_bufP;

   if ( ! bufP ) return;

   // this also reset the pull index
   ipiobuf_pull_packet_len(bufP);

   struct IppChunkh16 *firstcP = ipiobuf_pull_chunkh16(bufP);

   if ( ! firstcP ) return;

   switch ( firstcP->chunk_type )
      {
      case IPPHRchu_message_rh:
         prisoc_message_rh(socP, bufP);
         break;

      case IPPHRchu_tcp_rh_syn_cs:
         prisoc_parse_tcp_syn_cs(socP,bufP);
         break;

      case IPPHRchu_udp_data_rh:
         // there is nothing to do for UDP data
         // it is parsed at user level, way easier...
         break;

      default:
         fprintf(socP->debug_fd,"prisoc_parse_sock_reply: UNSUPP chunk_type=%d\n",firstcP->chunk_type);
      }
   }

/**
 * this will give the client an opportunity to know things have gone wrong
 */
static void prisoc_call_uparse(struct IPP_soc *const sockP )
   {
   if ( sockP->uparseFun )
      (*sockP->uparseFun)(sockP);
   }


/**
 * need to be called to retrieve a reply from the router
 * You can poll on the router FD for data to be read
 * NOTE: even if EOF is received, there MAY be a good buffer to be parsed
 * in theory, not, since it should read ALL the bytes, up to EOF, NOT including EOF
 * @return the number of bytes read or a NEGATIVE error code
 */
static int prisoc_read_router_reply(struct IPP_soc *const socP )
   {
   int errcode;

   if ( (errcode=prisoc_rx_pklen(socP)) )
      {
      // this will give the client an opportunity to know things have gone wrong
      prisoc_call_uparse(socP);
      return errcode;
      }

   int letti=0;

   while ( ! ipiobuf_rxnet_is_received(socP->rx_bufP) )
      {
      // even in local network read can be split into pieces
      errcode=ipiobuf_rxnet_from(socP->rx_bufP, socP->router_fd);

      if ( errcode < 0 )
         {
         // this is a read error, still, need to parse received buffer...
         ippsoc_close(socP);
         errcode=IPPS_ERR_EOF;
         break;
         }

      letti += errcode;
      }

   prisoc_parse_sock_reply(socP);
   prisoc_call_uparse(socP);

   return errcode < 0 ? errcode : letti;
   }


/**
 * this should connect to the router with a NEW connection and report that it is to service a previous incoming TCP syn
 * so, a NEW socket with the proper data...
 * https://man7.org/linux/man-pages/man2/accept.2.html
 * The info on last sequence and the cookies are in the listen socket !!
 * NOTE: this should copy the config from the listen sock and setup the new socket
 */
int ippsoc_accept(struct IPP_soc *listenP, struct IPP_soc *newsocP)
   {
   // Pick up the info on the incoming connection
   struct IPPS_tcp_accept_info *storP = listenP->acceptw_array;

   if ( ! storP->soface_idx )
      return prisoc_set_error(listenP,IPPS_ERR_BADSTATE,"ippsoc_accept: NO accept information");

   prisoc_init(newsocP,listenP);
   ippsoc_open(newsocP, NULL);

   if ( prisoc_isDebugMask(listenP, IPPSOC_DEBUG_ACCEPT))
      fprintf(listenP->debug_fd,"ippsoc_accept: soface=%d \n",storP->soface_idx);

   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(sizeof(struct IPPS_tcp_accept_info));

   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_cmd_accept_hr);

   struct IPPS_tcp_accept_info *reqP = ipiobuf_push_reserve(iobP, sizeof(struct IPPS_tcp_accept_info));

   // this is THE PARENT !!! this will allow the new socket to pick up some data from the parent, if needed
   reqP->soface_idx=storP->soface_idx;
   reqP->tcp_state=storP->tcp_state;   // I need to reply to the original address

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   // set this aside for debugging
   newsocP->tcp_state = storP->tcp_state;
   newsocP->tcp_state.tcpRole=SOCTCPROLE_server;

   int retcode;
   if ( (retcode=prisoc_tx_all_AA(newsocP,iobP,"accept")) )
      return retcode;

   // I can now safely clear up the element, to avoid possible reuse
   memzero_ptr(listenP->acceptw_array);

   // wait for the confirmation event
   if ( prisoc_isDebugMask(listenP, IPPSOC_DEBUG_ACCEPT))
      fprintf(listenP->debug_fd,"ippsoc_accept: soface=%d waiting for accept confirm \n",listenP->soface_idx);

   prisoc_qwait_timed(newsocP, IPPHRchu_make_naked_rh, newsocP->qw_timeout_s);

   if ( prisoc_isDebugMask(listenP, IPPSOC_DEBUG_ACCEPT))
      fprintf(listenP->debug_fd,"ippsoc_accept: soface=%d accept naked %d \n",listenP->soface_idx,newsocP->soc_naked);

   if ( newsocP->soc_naked )
      return newsocP->router_fd;
   else
      return prisoc_set_error(listenP,IPPS_ERR_TIMEOUT,"ippsoc_accept: timeout on accept");
   }

/**
 * this waits for an accept from some client and then behave like the standard accept
 */
int ippsoc_accept_wait(struct IPP_soc *listenP, struct IPP_soc *newsocP, uint16 wait_s)
   {
   // Pick up the info on the incoming connection
   struct IPPS_tcp_accept_info *storP = listenP->acceptw_array;

   if ( storP->soface_idx )
      return ippsoc_accept(listenP, newsocP);

   listenP->qw_have_event=0;  // make sure that possible old events are cleared
   prisoc_qwait_timed(listenP, IPPHRchu_tcp_rh_syn_cs, wait_s);

   if ( prisoc_isDebugMask(listenP, IPPSOC_DEBUG_ACCEPT))
      fprintf(listenP->debug_fd,"ippsoc_accept_wait: soface=%d accept naked %d \n",listenP->soface_idx,newsocP->soc_naked);

   // check again if anything has arrived
   if ( storP->soface_idx )
      return ippsoc_accept(listenP, newsocP);

   return prisoc_set_error(listenP,IPPS_ERR_TIMEOUT,"ippsoc_accept_wait: timeout on accept");
   }




static int prisoc_connect_udp(struct IPP_soc *const socP, struct IppAddress31 *toP, uint16 to_port )
   {
   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(100);

   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_connect_udp_hr);

   struct IPPS_hrp_connect_udp *reqP = ipiobuf_push_reserve(iobP, sizeof(struct IPPS_hrp_connect_udp));

   reqP->toippa = *toP;
   reqP->to_port = to_port;

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   return prisoc_tx_all_AA(socP,iobP,"UDP connect sent");
   }





/**
 * I need to prepare all params, since th check on actual good data will be done here, sock side
 * NOTE that if connect is successful I should make the socket naked
 * TCP datastream is ahndled like this, since it is the only way to pass an "opened" socket to an application
 * when doing fork-exec (like spawning a shell attached to a socket)
 */
static int prisoc_connect_tcp(struct IPP_soc *const socP, struct IppAddress31 *toP, uint16 to_port )
   {
   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(sizeof(struct IPPS_tcp_connect_hrp));

   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_connect_tcp_hr);

   struct IPPS_tcp_connect_hrp *reqP = ipiobuf_push_reserve(iobP, sizeof(struct IPPS_tcp_connect_hrp));

   reqP->soface_idx = socP->soface_idx;

   reqP->dest_ippa = *toP;
   reqP->dest_port = to_port;

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   socP->tcp_state.tcpRole=SOCTCPROLE_client;

   int retcode;
   if ( (retcode=prisoc_tx_all_AA(socP,iobP,"prisoc_connect_tcp sent")) )
      return retcode;

   // wait for the confirmation event
   if ( prisoc_isDebugMask(socP, IPPSOC_DEBUG_TCPCONNECT))
      fprintf(socP->debug_fd,"prisoc_connect_tcp: soface=%d waiting for connect confirm \n",socP->soface_idx);

   prisoc_qwait_timed(socP, IPPHRchu_make_naked_rh, socP->qw_timeout_s);

   if ( prisoc_isDebugMask(socP, IPPSOC_DEBUG_ACCEPT))
      fprintf(socP->debug_fd,"prisoc_connect_tcp: soface=%d connect naked %d \n",socP->soface_idx,socP->soc_naked);

   if ( socP->soc_naked )
      return socP->router_fd;
   else
      return prisoc_set_error(socP,IPPS_ERR_TIMEOUT,"prisoc_connect_tcp: timeout on connect");
   }

/**
 * attempt to connect to router and return 0 if all is fine or an error code and message
 * Basic address check is done but NO reachbility is done
 * So, the logic is that I connect to the router and tell it that I wish to do a UDP or TCP on the given port
 * TODO NOTE that for TCP this will ALSO make a connection naked !!!
 *
 * @param rtrP the router to connect to
 * @param sockP the datastructure to initialize
 */
int ippsoc_connect_A(struct IPP_soc *const socP, struct IppAddress31 *toP, uint16 to_port )
   {
   if ( ! socP ) return IPPS_ERR_NULLP;

   switch ( socP->soc_proto )
      {
      case ICHUNK_UDP:
         return prisoc_connect_udp(socP,toP,to_port);

      case ICHUNK_TCP:
         return prisoc_connect_tcp(socP,toP,to_port);

      default:
         return prisoc_set_error(socP,IPPS_ERR_BADPROTO,"connect: unsupported protocol");
      }
   }

int ippsoc_connect(struct IPP_soc *const socP, char *toP, uint16 to_port )
   {
   struct IppAddress31 to;
   ippsoc_string_to_address(&to, toP);
   return ippsoc_connect_A(socP,&to,to_port);
   }

/**
 * attempt to bind a given port to the socket, so, it is almost ready to be in listen mode
 * NOTE: the address is currently not really supported, you can safely pass NULL
 * The reality is that an interface does NOT have an address and therefore you cannot bind on that one
 * What could be done is binding on localhost (127) but I still need to write the code...
 */
int ippsoc_bind(struct IPP_soc *const socP, struct IppAddress31 *toP, uint16 to_port )
   {
   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(100);
   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_cmd_bind_hr);

   struct IPPS_hrp_bind *reqP = ipiobuf_push_reserve(iobP, sizeof(struct IPPS_hrp_bind));

   reqP->protocol = socP->soc_proto;
   reqP->port = to_port;

   // it is legal NOT to provide a binding address
   if ( toP ) reqP->ipp_address = *toP;

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   return prisoc_tx_all_AA(socP,iobP,"bind sent");
   }


int ippsoc_sendto(struct IPP_soc *socP, void *dataP, uint16 data_len, enum IppChunkid proto, struct IppAddress31 *toP, uint16 to_port)
   {
   if ( ! socP )
      return IPPS_ERR_NULLP;

   if ( ! dataP )
      return prisoc_set_error(socP,IPPS_ERR_NODATA,"missing data pointer");

   if ( ! toP )
      return prisoc_set_error(socP,IPPS_ERR_NODESTA,"missing destination address");

   uint16 w_buflen = data_len + sizeof(struct IPPS_hrp_udp_sendto)+10;

   if ( w_buflen > IPP_PACKET_DATALEN_MAX )
      return prisoc_set_error(socP,IPPS_ERR_TXLENBIG,"data len is too big");

   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(w_buflen);

   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_cmd_sendto_hr);

   struct IPPS_hrp_udp_sendto *reqP = ipiobuf_push_reserve(iobP, sizeof(struct IPPS_hrp_udp_sendto)+data_len);

   reqP->to_proto=proto;
   reqP->to_port=to_port;
   reqP->ipp_address=*toP;
   reqP->data_len=data_len;
   memcpy(reqP->data_array,dataP,data_len);

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   return prisoc_tx_all_AA(socP,iobP,"sendto");
   }


/**
 * NOTE: it is legal to send a zero length UDP packet !
 * The desired behavior is that you get an error if try to TX data too big
 */
static int prisoc_send_data_udp(struct IPP_soc *socP, void *dataP, uint16 data_len)
   {
   uint16 w_buflen = data_len + sizeof(struct IPPS_udp_data_hrp)+10;

   if ( w_buflen > IPP_PACKET_DATALEN_MAX )
      return prisoc_set_error(socP,IPPS_ERR_TXLENBIG,"data len is too big");

   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(w_buflen);

   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_udp_data_hr);

   struct IPPS_udp_data_hrp *reqP = ipiobuf_push_reserve(iobP, sizeof(struct IPPS_udp_data_hrp)+data_len);

   memcpy(reqP->data_array,dataP,data_len);

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   int errcode=prisoc_tx_all_AA(socP,iobP,"send UDP data");

   return errcode < 0 ? errcode : data_len;
   }


/**
 * public API
 * Send some data to router where it will be forwarded using a previous connect or bind
 * NOTE: the interface should be as much as possible like standard unix write
 * A difficult part will be to implement the poll interface, meaning, that I should have a way to
 * signal to the OS that there is no more space to write on this fd....
 * NOTE: it is legal to send a zero byte buffer using TCP, this could be to keep a connection alive
 * EOF is signaled using a specific error code !
 * @return the number of bytes written, of the original buffer or a negative number
 */
int ippsoc_write(struct IPP_soc *const socP, void *dataP, uint16 data_len )
   {
   if ( ! dataP )
      return prisoc_set_error(socP,IPPS_ERR_NODATA,"missing data pointer");

   if ( socP->soc_proto == ICHUNK_UDP )
      return prisoc_send_data_udp(socP, dataP, data_len );

//   if ( socP->soc_proto == ICHUNK_TCP )
//      return prisoc_send_data_tcp(socP, dataP, data_len );

   return prisoc_set_error(socP,IPPS_ERR_BADPROTO,"unsupported protocol");
   }

/**
 * write an ICMP request
 * This can be used to send different kind of ICMP request
 */
int ippsoc_write_icmp(struct IPP_soc *const sockP, void *dataP, uint16 data_len)
   {
   if ( ! dataP )
      return prisoc_set_error(sockP,IPPS_ERR_NODATA,"missing data pointer");

   uint16 w_buflen = data_len + sizeof(struct IPPS_udp_data_hrp)+10;

   if ( w_buflen > IPP_PACKET_DATALEN_MAX )
      return prisoc_set_error(sockP,IPPS_ERR_TXLENBIG,"data len is too big");

   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(w_buflen);

   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_icmp_hr);

   ipiobuf_push_array(iobP,dataP,data_len);

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   return prisoc_tx_all_AA(sockP,iobP,"send ICMP");
   }

/**

 When a UDP socket is created, its local and remote addresses are unspecified.
 Datagrams can be sent immediately using sendto(2) or sendmsg(2) with a valid destination address as an argument.
 When connect(2) is called on the socket, the default destination address is set and datagrams can now be sent using send(2) or write(2) without specifying a destination address.
 It is still possible to send to other destinations by passing an address to sendto(2) or sendmsg(2).
 In order to receive packets, the socket can be bound to a local address first by using bind(2).
 Otherwise, the socket layer will automatically assign a free local port out of the range defined by
 /proc/sys/net/ipv4/ip_local_port_range and bind the socket to
 */
int ippsoc_listen(struct IPP_soc *socP, uint8 backlog )
   {
   struct Ipiobuf * const iobP = ippsoc_new_ipiobuf(100);

   uint16 *clenP = ipiobuf_push_chunkx_header(iobP,IPPHRchu_cmd_listen_hr);

   struct IPPS_hrp_listen *reqP = ipiobuf_push_reserve(iobP, sizeof(struct IPPS_hrp_listen));

   reqP->backlog=backlog > IPPTCP_ACCEPTW_LEN ? IPPTCP_ACCEPTW_LEN : backlog;

   ipiobuf_set_chulen16(iobP,clenP);
   ipiobuf_set_pktlen_new(iobP);

   socP->tcp_state.tcpRole=SOCTCPROLE_server;

   return prisoc_tx_all_AA(socP,iobP,"listen sent");
   }


/**
 * Wrapper for the original inet_Aton to make it simpler and safer
 * If it fails the destination is zeroed
 * @return 0 if all is fine or an error code
 */
int unix_inetv4_aton (struct sockaddr_in *toP, char *srcP, uint16  port )
   {
   if ( toP == NULL ) return -1;

   memzero_ptr(toP);  // it is safe to zero it here

   if ( srcP == NULL )
      return -2;

   // if parsing is good the result will have this family
   toP->sin_family = AF_INET;
   toP->sin_port   = htons(port);

   // the original function returns non zero if the parsing is good
   if ( inet_aton(srcP, &toP->sin_addr) ) return 0;

   memzero_ptr(toP);

   return -3;
   }


/**
 * Convert an IPP address into a char buffer
 * the really full address may require up to 124 bytes
 * @return the zero terminated string or NULL if it does not fit or error
 */
char *ippsoc_address_to_string(char *toP, int to_len, struct IppAddress31 *addrP)
   {
   const int reserve_bytes=6;

   // if no destination is given, I return an empty string
   if ( ! toP ) return "";

   if ( to_len < reserve_bytes ) return "bufsmall";

   if ( ! addrP ) return "noaddrP";

   // clear up, makes easier to debug
   memset(toP,0,to_len);

   char * const risulP = toP;
   uint8 add_dot=FALSE;
   int read_index=0;

   char *ceilingP = toP + (to_len - reserve_bytes);   // do not write anything past this pointer

   while ( read_index < addrP->adr_len )
      {
      // If I have hit the ceiling, abort
      if ( toP >= ceilingP ) return "bufsmall";

      if (add_dot)
         *toP++ = '.';

      uint8 abyte=addrP->adr_v[read_index++];
      int written=snprintf(toP, reserve_bytes-1, "%d",abyte);

      if ( written <=0 )
         return NULL;

      toP = toP + written;
      add_dot=TRUE;
      }

   return risulP;
   }


/**
 * Convert a char buffer into an IPP address
 * NOTE that an empty address is somewhat possible, to indicate any possible address
 * @return the number of bytes in the address or zero if the address is empty
 */
uint8 ippsoc_string_to_address(struct IppAddress31 * const addrP, char * const fromP)
   {
   if ( ! addrP ) return 0;

   memzero_ptr(addrP);

   if ( ! fromP ) return 0;

   // need to copy it since it gets mangled by the strtok
   char fcopy[strlen(fromP)+2];
   strncpyda(fcopy,fromP,sizeof(fcopy));

   char *saveptr, *tokenP;
   uint8 a_len=0;

   tokenP = strtok_sep(fcopy, &saveptr, '.');

   while ( tokenP && (a_len < sizeof(addrP->adr_v)) )
      {
      int anint;

      if ( sscanf(tokenP,"%d",&anint) == 1 )
         addrP->adr_v[a_len++] = anint;

      tokenP = strtok_sep(NULL, &saveptr, '.');
      }

   addrP->adr_len=a_len;

   return a_len;
   }

/**
 * If you need to create a basic packet, you first call this one to allocate space
 * It does NOT need to be exact, just more than what is needed
 * It may be appropriate to use the biggest buffer, to avoid memory fragmentation
 * Or to use two sizes, 1024 and 2048 ...
 */
struct Ipiobuf * ippsoc_new_ipiobuf ( uint16 wish_size )
   {
   struct Ipiobuf * const risulP = ipiobuf_malloc(IPIOBUF_MAGIC_H, wish_size);

   ipiobuf_push_reserve(risulP,sizeof(uint16));  // for the packet length

   return risulP;
   }

/**
 * add the close on exec flag to the given descriptor
 */
void fcntl_set_close_on_exec ( int fd )
   {
   if ( fd < 0 ) return;

   int flags = fcntl(fd, F_GETFD);

   if ( flags == -1 ) return;

   flags |= FD_CLOEXEC;

   fcntl(fd, F_SETFD, flags);
   }
