/**
 * Copyright Damiano Bolla, Cologna Veneta, Italy, All right reserved
 * Eclipse does not have any more copyright manager, so I need to edit by hand
 * GNU LESSER GENERAL PUBLIC LICENSE Version 3, 29 June 2007
 */

#ifndef IPP_SOCK_H
#define IPP_SOCK_H

#include <hal.h>

#include <netinet/in.h>
#include <arpa/inet.h>

/**
 * since this is shared with userland, it is actually the baseline for the shared functions
 */

#define IPP_PACKET_DATALEN_MAX   (1518-2)    // bytes, it is the Ethernet frame size minus the two bytes packet header
#define IPP_ADDRESS_LEN_MAX      31          // bytes, NOTE that this PLUS the length makes the struct length EVEN


/**
 * Tested to allocate a uint8 because of STRUCT_PACKED
 * ICHUNK_GENERIC16 Packet should have a "variant" field to support protocol evolution

 * Note on the pletora of IP "protocols" like IPPROTO_IGMP, IPPROTO_IPIP, IPPROTO_GRE and so on
 * they ALL will be using the ICHUNK_GENERIC16, the chunk will be like this
 * uint8 chunk_id=ICHUNK_GENERIC16
 * uint8 extra info
 * uint16 chunk_len
 * what follows is a uint16 that defines WHAT is in the buffer
 * Remember that the chunk_len INCLUDES the length of the header, four bytes, that is
 *
 * NOTE: as much as possible AVOID using zero as a valid value, no index 0 is valid, NO configuration parameter is valid
 * the meaning of the value 0 MUST be equivalent to SQL NULL
 *
 * Let me try to have the LSB indicating a 16 bits len
 * So, the LSB bit of the ID indicates a 16bits chunk, with 16bits chunk_len !!!
 * You can therefore MIX chunk8 and chunk16 in a single packet !
 *
 * WARNING: there is a bit of confusion on the use of the chunkid as protocol
 * Quite convenient but it becomes a bit confusing with the 8 and 16 bit version of it
 * Generally speaking ALWAYS use the 8 bit version, the 16bit version is used ONLY to indicate the extra length !
 */
enum IppChunkid
   {
   ICHUNK_NULL=0,
   ICHUNK_ONEBYTE,         // this chunk id is used to indicate a one byte len chunk, to make chunks even start
   ICHUNK_IDENT_REQ,       // 8 bit ipface: if needed to login, this is the first that the HOST sends to the ROUTER
   ICHUNK_RES16_B,         // 16bit reserved, possible 16bits chunk
   ICHUNK_IDENT_RES,       // 8 bit ipface: Identity response, with some extra info, used to verify a link
   ICHUNK_RES16_C,         // 16bit reserved, possible 16bits chunk
   ICHUNK_LINKID,          // 8 bit radio,ipface: includes the ACK of a previous packet ALSO includes the magic for authentication
   ICHUNK_RES16_D,         // 16bit reserved, possible 16bits chunk
   ICHUNK_BEACON,          // 8 bit radio: a beacon is a periodic information that is not link related
   ICHUNK_RES16_E,         // 16bit reserved, possible 16bits chunk
   ICHUNK_HDR15,           // 8 bit route: IPP header with addresses as 15 bytes max, this can switch to 16 bits chunks !
   ICHUNK_RES16_F,         // 16bit reserved, possible 16bits chunk
   ICHUNK_HDR31,           // route: IPP header with addresses as 31 bytes max, this can switch to 16 bits chunks !
   ICHUNK_RES16_G,         // 16bit reserved, possible 16bits chunk
   ICHUNK_ICMP,            // 8bit route: This is a ICMP packet, includes routing info and whatever
   ICHUNK_ICMP16,          // 16bit size route: This is a ICMP packet, includes routing info and whatever
   ICHUNK_UDP,             // 8bit route: the payload is a udp16 packet
   ICHUNK_UDP16,           // 16bit route: the payload is a udp16 packet when the length is 16 bits
   ICHUNK_TCP,             // A TCP packet, when the length fits into a uint8
   ICHUNK_TCP16,           // TCP packet with length of 16 bits
   ICHUNK_GENERIC,         // A chunk holding a generic protocol, the first uint16 defines WHAT is the rest of packet
   ICHUNK_GENERIC16,       // Use this if the header is 16 bits
   } STRUCT_PACKED;



/**
 * this is a HOST representation of an address, NOT the one on the wire
 * as a general rule, the value 0 (zero) has a meaning of null
 * it should me possible to memset(ptr,0,sizeof(*ptr)) to clean up
 * You could do a strlen() on adr_v and it should be the same as adr_len
 * NOTE that it is ASSUMED that the size of this struct is EVEN
 */
struct IppAddress31
   {
   uint8 adr_len;                     // this is the LEN of the address
   uint8 adr_v[IPP_ADDRESS_LEN_MAX];  // adr_v[0] is the MSB byte of the address, normally should be value 1
   };


/**
 * Very useful to have a pointer to the chunk header
 * In a IPP packet you HAVe to begin with a chunk8 and then the address MAY decide that following chunks are chunk16
 * NOTE that internally, chunk16 is a superset of chunk8, so, it is kind of legal to extract everything to Chunk16
 */
struct IppChunkh8
   {
   enum IppChunkid chunk_type;
   uint8 chunk_len;
   };

/**
 * The field chunk_variant can be used in different ways in different context
 * Eg: Use to define what "subprotocol" we are wrapping, if you need it, eg: IPPROTO_ESP
 * It is up to the sender of a packet to decide to use Chunk16 or not
 */
struct IppChunkh16
   {
   uint8  chunk_type;      // to assign the right type I have to adjust the sock part..
   uint8  chunk_variant;   // try not to use it....
   uint16 chunk_len;       // the TOTAL length of the chunk, including the header !
   };

/**
 * Most compact way to send a version like 2024050518, instead of bytes
 */
union IppVersion
   {
   struct
      {
      uint32 hour:5;       // 0 to 31
      uint32 mday:5;       // 0 to 31
      uint32 month:4;      // 0 to 15
      uint32 year:18;      // 0 to 262143 MSB on LE CPU
      } bits;

   uint32 val;
   } STRUCT_PACKED;


/**
 * This is by definition, localhost
 * You can use all the space below it, just this is localhost
 * At ANY level, sending something to 127 is sending it to localhost
 * BUT an address like 127.1.127.127 should be legal and routable
 * the key point is that it is an address of length 1 and value 127
 */
#define IPP_LOCAL_ADDR  "127"

/**
 * this is the public part of UDP handling
 */
#define IPPUDP_PARROT_PORT    10    // it just reply with a parrot packet, most like ICMP PING
#define IPPUDP_GPSDB_PORT     11    // the port where the gpsdb will be "listening"


#define IPP_ERRMSG_LEN     100



/**
 linkid has its own chunk and it is sent on links that are "shared" or need identification
 NOTE that vmagic is made using the secret + the router seed + the remaining packet
 the idea is that the sequence allows to retransmit even partial packets
 before considering a packet received I have to receive the ACK to MY sequence

 pkt_vmagic is sent by router and client when there is a linkid
 a linkid is created ONLY when there is an IDENTIFY req - res and hence there is a shared secret between router and station
 the idea is that the the sender zero vmagic, calculate the hash putting the secret at the beginning and then send the resulting last byte

 the receiver does the same with his own secret and see if the magic is the same
 this, of course, after checksum is valid ....

 A router to router will send a bunch of packets each giving the full source and destination
 it then also adds the linkid and vmagic

 NOTE So, assume router sends data to sta, sta receive and send ack, but get lost
 router RESEND the packet and I have a duplicate packet...
 So, yes, I need a sequence and when I rx I have to rx something that is > previous
 NOTE: If I RX something I should IN ANY case ACK it even if I discard

 NOTE: Careful, if I am sending big chunks, the header of Linkid will be 16 bits !
 */

struct IppLinkid
   {
   // this structure is the one sent over the network  ICHUNK_LINKID
   uint16 pkt_liface_idx;  // the ipface this connection is bound to, NOTE that depending on the "master" it may refer to different "objects" !!!
   uint16 pkt_vmagic;      // explained above, a way to "certify" a packet
   uint8  pkt_seq;         // this packet sequence, the sequence of THIS packet
   uint8  ack_rx_seq;      // linkid acknowledge the RX of the given sequence
   };


/**
 * ================================================== SOCKET SPECIFIC
 * in IPP a write will return the number of bytes written, INCLUDING 0 if nothing has been given to write OR cannot write anything
 * it is NOT an EOF, EOF has a well defined error code, negative !
 * this is ALSO for similarity with UDP where you CAN send a zero length packet !
 */

#define IPPS_DONE           0        // Unless noted differently, this means operation completed
#define IPPS_ERR_EOF        -1       // EOF is this error code, NOT 0
#define IPPS_ERR_NULLP      -2       // the given pointer is NULL
#define IPPS_ERR_NO4SO      -3       // Cannot create a ipv4 socket to connect to router
#define IPPS_ERR_BADV4      -4       // Bad ipv4 address
#define IPPS_ERR_BSIZEOF    -5       // Bad sizeof key structure
#define IPPS_ERR_SOCKCLO    -6       // Socket is closed
#define IPPS_ERR_NORXLEN    -7       // Could not get a receive length
#define IPPS_ERR_NOMALLOC   -8       // Memory allocation fail
#define IPPS_ERR_NULLPORT   -9       // No port given when required
#define IPPS_ERR_PORTIUSE   -10      // Given port is in use
#define IPPS_ERR_TXLENBIG   -11      // too much data to send
#define IPPS_ERR_NODESTA    -12      // no destination address when required
#define IPPS_ERR_NODATA     -13      // no data given when required
#define IPPS_ERR_BADPROTO   -14      // Given protocol is invalid
#define IPPS_ERR_WRITE4FAIL -15      // IPV4 write fail
#define IPPS_ERR_OPALDONE   -16      // The required operation has already been done
#define IPPS_ERR_TXOVERRUN  -17      // Attempting to TX twice on same socket
#define IPPS_ERR_BADSTATE   -18      // Request is not compatible with socket state
#define IPPS_ERR_NOCONN     -19      // Cannot connect to router or something else
#define IPPS_ERR_TIMEOUT    -20      // timeout while waiting for something


/**
 * this is how you reach the router, at the moment, trough IPv4
 * data between the application and the router will be chunked, meaning that every "block" will have
 * uint16 length
 * uint8 chunk_id
 * uint8 variant
 * remaining data
 * The same logic of OOB data apply, you CANNOT jump the queue but you can send OOB data
 * so, you can just block on read/write or use poll
 * the key point is that you MUST call the IPP_read IPP_write so they can unpack the data coming from the stream in the appropriate way
 */


/**
 * This is used by the API, internally, to transfer data between the application and the router
 * Host Router chunk ID
 * NOT PUBLIC INTERFACE, NOT used to route packets in public network
 * Since application WILL have this codes compiled in, you will NOT be able to "renumber" them at will
 * NOTE: close is mostly unneeded since sock will be closed when connection TCP is closed
 */
enum IppHRchunkid
   {
   IPPHRchu_null=0,
   IPPHRchu_ipv4_connect_hr,  // will be sent when the initial ipv4 connection is sent
   IPPHRchu_ipv4_connect_rh,  // response
   IPPHRchu_connect_udp_hr,   // UDP and TCP are separate since they have vastly different params
   IPPHRchu_connect_tcp_hr,   // TCP requires some cookie and a sequence number
   IPPHRchu_cmd_connect_rh,   // response
   IPPHRchu_cmd_listen_hr,    // request to listen on a port, a bind is NOT enough !
   IPPHRchu_cmd_listen_rh,    // response
   IPPHRchu_cmd_accept_hr,    // an incoming TCP connection can be accepted and result into a new socket
   IPPHRchu_cmd_accept_rh,    // response
   IPPHRchu_cmd_sendto_hr,    // sendto allows to send UDP
   IPPHRchu_cmd_sendto_rh,    // response, if needed
   IPPHRchu_udp_data_hr,      // host to router data, just data, can be used FROM router to indicate error situation
   IPPHRchu_udp_data_rh,      // router to host
   IPPHRchu_oobdata_hr,       // host to router data OOB, sending data
   IPPHRchu_oobdata_rh,       // router to host OOB
   IPPHRchu_cmd_bind_hr,      // bind an address to this socket
   IPPHRchu_cmd_bind_rh,      // response
   IPPHRchu_cmd_shutdown_hr,
   IPPHRchu_cmd_shutdown_rh,
   IPPHRchu_cmd_close_hr,     // close the port at router side
   IPPHRchu_cmd_close_rh,     // response
   IPPHRchu_icmp_hr,          // host to router ICMP message, well, this is actually a PING req !
   IPPHRchu_icmp_rh,          // router to host ICMP, this is more likely to happen
   IPPHRchu_message_rh,       // used by router to send a generic errcode to application, when one of the previous one do not apply
   IPPHRchu_tcp_rh_syn_cs,    // router inform socket that a SYN client -> server has been received
   IPPHRchu_make_naked_rh,    // router inform socket that a router connection is now naked
   };

// ================================================== PRIVATE IPP, NOT public API

#define IPPS_hrs_header_LEN 4  // used to verify if compiler knows sizeof() correct even for enum

/**
 * When a router just want to send a generic response it will use this data
 * NOT PUBLIC INTERFACE
 */
struct IPPS_hrp_reply
   {
   int  hrp_rescode;                  // can be a negative number, generally speaking 0 means all is fine
   uint16 soface_idx;                 // this is the soface bound to this response
   char hrp_message[IPP_ERRMSG_LEN];  // a message to report to user
   enum IppHRchunkid o_request;       // the original request that generated this response message
   };

/**
 * parameters to include in a connection request from host to router
 * NOT PUBLIC INTERFACE
 */
struct IPPS_hrp_connect_udp
   {
   uint16 to_port;
   struct IppAddress31 toippa;
   };

/**
 * parameters to include in BIND request from application to router
 * in IPP a router can have only ONE address, a host has an address regardless of all the possible interfaces
 * so, when you BIND you can only bind for yourself or nothing
 * NOT PUBLIC INTERFACE
 */
struct IPPS_hrp_bind
   {
   uint16 port;               // the classic meaning of an IP port
   enum IppChunkid protocol;  // one of ICHUNK_UDP, ICHUNK_TCP, ICHUNK_GENERIC_A
   struct IppAddress31 ipp_address;  // mostly  useless, but we will see
   };

/**
 * NOTE: if you want to send from a specific port, you need to bind before calling sendto
 */
struct IPPS_hrp_udp_sendto
   {
   enum IppChunkid to_proto;
   uint8  align_even;
   uint16 to_port;
   struct IppAddress31 ipp_address;   // send to this address
   uint16 data_len;        // kind of duplicate but makes code much simpler
   uint8  data_array[0];   //
   };

struct IPPS_udp_data_hrp
   {
   uint8  data_array[0];   //
   };

struct IPPS_udp_data_rhp       // Router to Host Packet, this is after the Chunk16 header and is ? part of the data"
   {
   struct IppAddress31 from_address;  // send to this address
   uint16 from_port;                  // a possible server, receiving a packet, need to know the client port to reply
   uint8  data_array[0];              // from here is the rest of data
   };

struct IPPS_hrp_listen      // host to router, to pass the listen options
   {
   uint8  backlog;
   };

// ======================================================== TCP ==========================================================

/**
 * The key point is that quite a few of the original bits are somewhat incompatible, like having a SYN + RST
 * So, the proper way is to have a state machine and well identified codes to be sure on where is data going
 * So, this follows the logic of TCP but should be way safer
 * this applies to the PACKET, meaning, what is the state that a PACKET can be
 * Server possible states are DIFFERENT from packet and for clients !!!
 */
enum IppTcppLinkState
   {
   IPPTCP_null,         // zero is NOT a valid packet state
   IPPTCP_SYN_cs,       // SYN sent from client to server
   IPPTCP_SYN_sc,       // SYN sent from server to client
   IPPTCP_ACK_cs,       // Client acknowledge the SYN sent by the server, from now on the connection is alive
   IPPTCP_DATA_cs,      // data client -> server
   IPPTCP_DATA_sc,      // data server -> client
   IPPTCP_FIN_cs,       // client request a FIN to server
   IPPTCP_FIN_sc,       // server request a FIN to client
   IPPTCP_RST_cs,       // client request a reset to server
   IPPTCP_RST_sc,       // server request a reset to client
   } STRUCT_PACKED;


#define TCP_MISSED_LEN  9         // should be in such a way that it makes the packet even


enum IppsocTcpRole
   {
   SOCTCPROLE_null,
   SOCTCPROLE_client,
   SOCTCPROLE_server,
   };


// https://en.wikipedia.org/wiki/Transmission_Control_Protocol
enum SoccliTcpState
   {
   SOTCPC_NULL,
   SOTCPC_SYN_SENT_cs,
   SOTCPC_SYNACK_SENT,
   SOTCPC_SYN_RECV,
   SOTCPC_ESTABLISH,
   SOTCPC_FINW1,        // FIN WAIT 1
   SOTCPC_FINW2,
   SOTCPC_CLOSEW,
   SOTCPC_CLOSING,
   SOTCPC_LASTACK,
   SOTCPC_TIMEW,
   SOTCPC_CLOSED,
   };

#define IPPTCP_ACCEPTW_LEN    2    // max number of accept that can wait on a single socket
#define IPPTCP_SENTLEN        10

/**
 * this capture the state information of a TCP socket
 * NOTE: You need a sent array to store the packets that have been sent but not yet confirmed
 * the cookie is used ONLY in the initial SYNC client -> server -> client
 */
struct IppTcpState
   {
   enum IppsocTcpRole    tcpRole;      // I need to know if this is server side or client side
   enum SoccliTcpState   tcp_sostate;    // this will be different depending on server or client !
   enum IppTcppLinkState tcp_listate;  // logically, almost classic TCP flags

   struct IppAddress31 dest_ippa;    //
   uint16 dest_port;       // kind of overlap what is in the server soface ?
   uint8  this_pk_seq;     // the packet sequence to send, once sent it is incremented by one
   uint8  last_rxed_seq;   // the sequence that I last received
   uint16 mss_len;         // the current mss length, the actual number
   uint8  window_len;      // max number of PACKETS that can be sent before stop sending (from peer latest_rx_seq)
   uint8  connect_cookie[TCP_MISSED_LEN];   //
   uint8  missing_packets[TCP_MISSED_LEN];  // store missing packets here !
   };

/**
 * NOTE that this is OK to be sent even for initial dialog, basically, the SOCKET size fill up most of data
 * this MUST be saved and will be used to RESEND the data if the other side say that it missed a packet
 * Initially, I could just have an array, the KEY point is to have the my_packet_seq
 * ALSO if I resend the data BUT the missed changes, I should send the CHANGED list of packets missed !!
 */
struct IPPS_tcp_connect_hrp      // host to router
   {
   uint16 soface_idx;         // Used in different ways, depending on the context
   struct IppAddress31 dest_ippa;
   uint16 dest_port;
   };

/**
 * Used to send info on the accept TO che APP that is listening and BACK to the router to complete the accept
 */
struct IPPS_tcp_accept_info        // router to host,
   {
   uint16 soface_idx;                 // this is the soface that did manage this tcp data
   struct IppTcpState tcp_state;      // the state as reported by the server
   };


/**
 * =================================================== constants for ICMP handling, mostly PING
 */

/**
 * This is kind of shared with the socket part, meaning that we have the same codes
 */
enum IppIcmpCode
   {
   ICMP_CODE_UNREACHABLE=0x03,
   ICMP_CODE_PING=0x08,
   } STRUCT_PACKED;

/**
 * IPP ping report with a message if a destination is unreachable
 * This is also shared with the application code
 */
enum IppIcmpPingReason
   {
   ICPING_REQ=1,
   ICPING_RES_REACH,
   ICPING_RES_UNREACH,   // this will be when the given port or router is unreachable
   } STRUCT_PACKED;

/**
 * Shared definition of unreachable reason
 */
enum IppIcmpUnreachReason
   {
   ICUNREACH_ADDRESS=1,
   ICUNREACH_TTL,
   ICUNREACH_PORT
   } STRUCT_PACKED;

/**
 * this is for ICMP PING request
 * The data that an application and the "kernel" exchange is this, different than what is on the net
 * the icmp_code is a simple way to differentiate between the various ICMP
 */
struct IPPS_hri_ping_req
   {
   enum IppIcmpCode icmp_code;
   uint8  to_protocol;  // can be 0 to indicate the first available protocol
   uint8  initial_ttl;  // so, it is possible to trigger the unreachable  TODO
   struct IppAddress31 to_address;
   uint16 to_port;      // can be zero to indicate "the router"
   };

struct IPPS_rhi_ping_res
   {
   enum IppIcmpCode icmp_code;
   enum IppIcmpPingReason ping_reason;
   struct IppAddress31 from_address;
   uint8  from_protocol;  // can be 0 to indicate the first available protocol
   uint16 from_port;      // can be zero to indicate "the router"
   };

struct IPPS_rhi_unreach_res
   {
   enum IppIcmpCode icmp_code;
   enum IppIcmpUnreachReason unreach_reason;
   struct IppAddress31 o_to_address;   // original destination address
   uint8  o_to_protocol;  // original to protocol
   uint16 o_to_port;      // original to port
   };

/**
 * this capture the fact that all response for ICMP have the icmp_code at the beginning
 */
union IPPS_rh_icmp_all
   {
   struct IPPS_rhi_ping_res ping_res;
   struct IPPS_rhi_unreach_res unreach_res;
   enum IppIcmpCode icmp_code;
   };


/**
 * ==================================== END of ICMP ===========================================
 */


struct IPP_soc;

/**
 * If you call ippsoc_read_reply() you can pass a function that will be used to callback and parse the data BEFORE freeing it
 * if you do NOT need to do local parsing , just pass NULL
 */
typedef void (*IppsocRxparseFun)(struct IPP_soc *socP);


/**
 * note that queuing will work at router level, I just need one buffer, here
 * now, some application will happly fork a client socket to a shell,so, there has to be a way to
 * provide a standard IO to them, it is not enough to provide an API
 * the solution I see is to make the connection to the router naked and let the router handle the TCP protocol
 * basically, all TCP params should be given to the router and the connection will become just a stream
 */

struct IPP_soc
   {
   pthread_t read_from_router_thread;  // At the moment, I make it quick wi h two threads
   pthread_mutex_t qw_mutex;        // mutex to guard queue access
   pthread_cond_t  qw_cond;         // used to wait on data
   IppsocRxparseFun uparseFun;      // if defined a user parse function to call when data arrives
   struct sockaddr_in rtrc_addr;    // the connection to the router, includes the port
   uint16 soface_idx;               // for debug AND to complete the TCP accept
   uint16 soc_from_port;            // the port data are from
   uint16 soc_to_port;              // the port I received on open
   uint16 qw_timeout_s;             // number of seconds for queue operations timeout
   uint32 debug_mask;               //

   FILE  *debug_fd;                 // you can redirect debug where you wish

   int router_fd;                   // talking to the router, here, can be polled

   struct Ipiobuf *tx_bufP;         // write buffer, data that goes from APP to router
   struct Ipiobuf *rx_bufP;         // read buffer

   int soc_errcode;                 // it it fails this is an errcode
   char soc_errmsg[IPP_ERRMSG_LEN]; // a possible error message

   int rtr_errcode;
   char rtr_errmsg[IPP_ERRMSG_LEN]; // a possible error message

   struct IppTcpState tcp_state;     // needed to manage TCP
   struct IPPS_tcp_accept_info acceptw_array[IPPTCP_ACCEPTW_LEN];   // received packets for SYNC request come here

   enum IppChunkid   soc_proto;     // the socket protocol
   enum IppHRchunkid qw_have_event; // the signal will set the received event here
   uint8 router_eof;                // if true then the router has EOF
   uint8 soc_naked;                 // if true the socket is naked and the service thread should exit
   };

int ippsoc_init(struct IPP_soc *const sockP, enum IppChunkid proto, char *rtr_ipv4, uint16 rtr_port, uint32 debug_mask);
int ippsoc_open(struct IPP_soc *const socP, IppsocRxparseFun funP);

int ippsoc_print_stats(struct IPP_soc *const sockP);
int ippsoc_set_uparsefun(struct IPP_soc *const socP, IppsocRxparseFun funP);

/**
 * If you need to know is a given SOCK is connected to the router
 */
int ippsoc_is_ipv4connect(struct IPP_soc *const sockP);

int ippsoc_connect(struct IPP_soc *const socP, char *toP, uint16 to_port );
int ippsoc_connect_A(struct IPP_soc *const socP, struct IppAddress31 *toP, uint16 to_port );

int ippsoc_bind(struct IPP_soc *const socP, struct IppAddress31 *toP, uint16 to_port );

int ippsoc_listen(struct IPP_soc *sockP, uint8 backlog );
int ippsoc_accept(struct IPP_soc *listenP, struct IPP_soc *dataP);
int ippsoc_accept_wait(struct IPP_soc *listenP, struct IPP_soc *newsocP, uint16 wait_s);

/**
 * This is IPP write, if you wish to have "classic" stream socket get the naked one
 */
int ippsoc_write(struct IPP_soc *const sockP, void *dataP, uint16 data_len );

int ippsoc_write_icmp(struct IPP_soc *const sockP, void *dataP, uint16 data_len);

int ippsoc_close (struct IPP_soc *sockP );

/*
 * ===========================================================================
 * The following are here since they are used by router AND app API
 */

int unix_inetv4_aton (struct sockaddr_in *toP, char *srcP, uint16  port );
void fcntl_set_close_on_exec ( int fd );

char *ippsoc_address_to_string(char *toP, int to_len, struct IppAddress31 *addrP);
uint8 ippsoc_string_to_address(struct IppAddress31 *addrP, char *fromP);
struct Ipiobuf * ippsoc_new_ipiobuf ( uint16 wish_size );





#endif
