/**
 * Copyright Damiano Bolla, Cologna Veneta, Italy, All right reserved
 * Eclipse does not have any more copyright manager, so I need to edit by hand
 * GNU LESSER GENERAL PUBLIC LICENSE Version 3, 29 June 2007
 */

#include <stringd.h>
#include <ctype.h>
#include <string.h>


/**
 * Copies at most dest_len-1 chars and ensure that the dest is zero terminated
 * destP will NOT be overwritten if params are BAD
 * @destP NOT NULL destination
 * @dataP NOT NULL source
 * @dest_len >= 1
 * @return a pointer where you could eventually append more data
 */
char *strncpyda ( void *toP, void *srcP, int dest_len )
   {
   uint8 *ctoP = toP;
   uint8 *csrcP = srcP;

   char *risulP = toP;

   if ( dest_len < 1 )
      {
      // there is not even the space for the termination zero
      return toP;
      }

   // one space reserved for the zero termination
   dest_len--;

   while ( (dest_len-- > 0) && (*ctoP++ = *csrcP++) ) risulP++;

   // ensure string is zero terminated
   *ctoP++ = 0;

   return risulP;
   }



/**
 * Convert the given integer with fixed decimals into a string
 * the resulting string MUST be zero terminated
 */
char *string_int32dec_to_string(char *const toP, int to_len, int32 int32dec, uint8 const decimals )
   {
   if ( ! toP ) return "null";

   if ( to_len < 1 ) return "(null)";

   memset(toP,0,to_len);  // clear all up so I do not have to worry

   int remain_space=to_len-1;

   char *wrP = toP;

   if ( remain_space > 0 && int32dec < 0 )
      {
      *wrP++ = '-';
      remain_space--;
      int32dec = - int32dec;
      }

   int divisor = 1;
   int counter=decimals;
   while ( counter-- > 0 ) divisor *= 10;

   int msb = int32dec / divisor;

   if ( remain_space > 2 )
      {
      int written=snprintf(wrP,remain_space,"%d",msb);
      if ( written < 0 || written > remain_space )
         return toP;

      wrP += written;
      remain_space -= written;
      }

   int decival=int32dec - (msb*divisor);

   if ( remain_space > 2 && decival > 0 )
      {
      *wrP++ = '.';
      remain_space--;

      int written=snprintf(wrP,remain_space,"%0*d",decimals, decival);
      if ( written < 0 || written > remain_space )
         return toP;

      wrP += written;
      remain_space -= written;
      }

   return toP;
   }



char *strcpyda ( char *destP, char *dataP )
   {
   char *risulP = destP;

   while ( (*destP++ = *dataP++) ) risulP++;

   return risulP;
   }



/**
 * copy a string into a limited space and also remove the escapes that have
 * been applied to tab and \n or \r
 * So any \t or \r or \n will be converted to the correct char
 * Copies at most dest_len-1 chars and ensure that the dest is zero terminated
 * destP will NOT be overwritten if params are BAD
 * @destP NOT NULL destination
 * @dataP NOT NULL source
 * @dest_len >= 1
 * @return a pointer where you could eventually append more data
 */
char *strncpy_escape ( char *destP, char *dataP, int dest_len )
   {
   char *risulP = destP;
   uint8 is_quoted = 0;
   char ach;

   if ( dest_len < 1 )
      {
      // there is not even the space for the termination zero
      return destP;
      }

   // one space reserved for teh zero termination
   dest_len--;

   while ( (ach = *dataP++) )
      {
      if ( is_quoted )
         {
         if ( ach == '\\' )
            ach = '\\';
         else if ( ach == 't')
            ach = '\t';
         else if ( ach == 'n')
            ach = '\n';
         else if ( ach == 'r')
            ach = '\r';

         is_quoted = 0;
         }
      else
         {
         if ( ach == '\\' )
            {
            is_quoted = 1;
            continue;
            }

         // if I am not quoting and the char is not the quoter starter, just go on
         }

      // if there is no space to insert (rememeber that a byte is already taken)
      if ( dest_len-- <= 0 ) break;

      *destP++ = ach;

      risulP++;
      }

   // ensure string is zero terminated
   // note how risulP is not incremented so it points right over the zero
   *destP++ = 0;

   return risulP;
   }

/**
 * escape the given srcP into destP qith a max of destLen
 * The escaped chars are \t \r \n
 * If there is no enough space the risul will be truncated
 * @return the given destP so you can concatenate the call into a printf
 */
char *string_escape (char *destP, int destLen, char *srcP )
   {
   char *risulP = destP;
   char ach;

   if ( destP == NULL ) return NULL;

   if ( destLen < 1 ) return destP;

   if ( srcP == NULL )
      {
      memset(destP,0,destLen);
      return destP;
      }

   // leave space for the terminating char at the end
   destLen--;

   while ( (destLen-- > 0) && (ach = *srcP++) )
      {
      if ( ach == '\t' )
         {
         destLen--;  // I know I have to use an extra char
         if ( destLen < 1 ) break;  // if I cannot fit it then get out

         *risulP++ = '\\';
         *risulP++ = 't';
         }
      else if (ach == '\r')
         {
         destLen--;  // I know I have to use an extra char
         if ( destLen < 1 ) break;  // if I cannot fit it then get out

         *risulP++ = '\\';
         *risulP++ = 'r';
         }
      else if (ach == '\n')
         {
         destLen--;  // I know I have to use an extra char
         if ( destLen < 1 ) break;  // if I cannot fit it then get out

         *risulP++ = '\\';
         *risulP++ = 'n';
         }
      else
         {
         *risulP++ = ach;
         }
      }

   *risulP = 0;  // insert terminating char

   return destP;
   }


/**
 * Attempt to convert the given string int a uint32 value
 * @return the new value of def_value on oerror
 */
uint32 string_HexToUint32 ( char *srcP, uint32 def_value )
   {
   uint32 aval;

   if ( ! srcP ) return def_value;

   if ( sscanf(srcP,"%x",&aval) != 1 ) return def_value;

   return aval;
   }


/**
 * Convert the HEX string to a uint32
 * It is an INTEGER conversion, therefore, it stops at ANY trailing . or space
 * @return TRUE if it has found at least one digit
 */
uint8 string_HexToUint32params ( struct string_HexToUint32T *paramsP )
   {
   if ( paramsP == NULL ) return FALSE;

   paramsP->risul=0;  // clean up

   if ( paramsP->strP == NULL ) return FALSE;

   paramsP->strP = string_strip_leading_space(paramsP->strP);

   if ( string_starts_with(paramsP->strP,"0x") )
   	{
		// IF the string starts with 0x then skip it away
   	paramsP->strP++;
		paramsP->strP++;
   	}

   uint32 risul = 0;
   uint8 found_digit=FALSE;
   uint8 ch;

   while ( (ch=*paramsP->strP++) )
      {
      if ( ch == ' ' || ch == '.' )
         {
         if ( found_digit )
            {
            // exit at the first non digit
            break;
            }
         else
            {
            // skip any leading spaces
            continue;
            }
         }

      risul = risul << 4;
      risul = risul + string_DigitToNibble(ch);

      found_digit = TRUE;
      }

   paramsP->risul = risul;

   return found_digit;
   }


/**
 * Convert the string to a decimal 32 bits
 * It is an INTEGER conversion, therefore, it stops at ANY trailing . or space
 * @return TRUE if it has found at least one digit, FALSE otherwise
 */
uint8 string_DecimalToInt32 ( struct string_DecimalToInt32T *paramsP )
   {
   int32 risul = 0;
   uint8 found_digit=FALSE;
   uint8 is_negative=FALSE;
   char ch;

   if ( paramsP == NULL ) return FALSE;

   paramsP->z_count=0;
   paramsP->risul=0;

   if ( paramsP->strP == NULL ) return FALSE;

   while ( (ch=*paramsP->strP++) )
      {
      if ( ch == '-' )
         {
         is_negative = TRUE;
         continue;
         }


      if ( ch == ' ' || ch == '.' || ch == '+' )
         {
         if ( found_digit )
            {
            // exit at the first non digit
            break;
            }
         else
            {
            // skip any leading spaces, dots or plus
            continue;
            }
         }

      // this can be done regardless of how many zeros are in front
      risul = risul * 10;
      int digitval = ch - '0';
      risul = risul + digitval;

      if ( digitval==0 && risul==0 )
         paramsP->z_count++;

      found_digit = TRUE;
      }

   if ( is_negative ) risul = -risul;

   paramsP->risul = risul;

   return found_digit;
   }

/**
 * Convert the string to a decimal 32 bits WITH decimals !!
 * @return TRUE if it has found at least one digit, FALSE otherwise
 */
uint8 string_DecimalToInt32dot ( struct string_DecimalToInt32T *paramsP )
   {
   int32  risul_int = 0;   // this is the integer part of the result
   uint32 risul_dec=0;     // this is the DECIAML part of the result

   uint8 found_digit=FALSE;
   uint8 is_negative=FALSE;
   uint8 found_dot=FALSE;

   if ( paramsP == NULL ) return FALSE;

   paramsP->z_count=0;
   paramsP->risul=0;

   if ( paramsP->strP == NULL ) return FALSE;

   char ch;
   while ( (ch=*paramsP->strP++) )
      {
      if ( ch == '-' )
         {
         is_negative = TRUE;
         continue;
         }

      if ( ch == '.' )
         {
         found_dot=TRUE;
         continue;
         }

      if ( ! isdigit(ch) )
         {
         if ( found_digit )
            break;
         else
            continue;
         }

      int digitval = ch - '0';

      if ( found_dot )
         {
         // this handles the decimal part
         risul_dec = risul_dec * 10 + digitval;
         }
      else
         {
         // this handles the integer part, before the dot
         risul_int = risul_int * 10 + digitval;
         }

      if ( digitval==0 && risul_int==0 )
         paramsP->z_count++;

      found_digit = TRUE;
      }

   uint8 decimals=paramsP->w_decimals;

   while (risul_int > paramsP->w_max_int )
      risul_int = risul_int / 10;

   while (decimals-- > 0)
      risul_int = risul_int * 10;

   risul_int = risul_int + risul_dec;

   if ( is_negative ) risul_int = -risul_int;

   paramsP->risul = risul_int;

   return found_digit;
   }



/**
 * Convert a int32 value (a SIGN value to a decimal representation
 * @param dest pointer to char array where to write the result must be at least 12 digits long
 * @param value the value to convert
 * @return an updated pointer to buffer where you can insert further data
 */
char *string_int32ToDecimal(char *dest, int32 value)
   {
   unsigned char digits[10];
   unsigned char digit_index = 9;

   if ( value == 0 )
      {
      *dest++ = '0';
      return dest;
      }

   memset(digits,0,sizeof(digits));

   if ( value < 0 )
      {
      *dest++ = '-';
      value = -value;
      }

   while ( value > 0 )
      {
      long resto = value / 10;
      unsigned char digit = value - resto * 10;
      digits[digit_index--] = '0'+digit;
      value = resto;
      }

   for ( digit_index++; digit_index<sizeof(digits); digit_index++)
      *dest++ = digits[digit_index];

   return dest;
   }


/**
 * Convert a uint32 value to a decimal representation
 * @param dest pointer to char array where to write the result must be at least 12 digits long
 * @param value the value to convert
 * @return an updated pointer to buffer where you can insert further data
 */
char *string_uint32ToDecimal(char *dest, uint32 value)
   {
   unsigned char digits[10];
   unsigned char digit_index = 9;

   if ( value == 0 )
      {
      *dest++ = '0';
      return dest;
      }

   memset(digits,0,sizeof(digits));

   while ( value > 0 )
      {
      long resto = value / 10;
      unsigned char digit = value - resto * 10;
      digits[digit_index--] = '0'+digit;
      value = resto;
      }

   for ( digit_index++; digit_index<sizeof(digits); digit_index++)
      *dest++ = digits[digit_index];

   return dest;
   }



/**
 * Convert a uint32 value to a hex representation
 * NO 0x will be inserted at the beginning !!!
 * @param dest pointer to char array where to write the result must be at least 12 digits long
 * @param value the value to convert
 * @param append if TRUE it will return a pointer where to add firther data
 * @return an updated pointer to buffer where you can insert further data
 */
char *string_uint32ToHex(char * const toP, uint32 value, uint8 append )
   {
   char *destP = toP;

   unsigned char digits[12];
   unsigned char digit_index = 11;

   memset(digits,0,sizeof(digits));

   do
      {
      digits[digit_index--] = string_nibbleToHex(value & 0x0F);
      value = value >> 4;
      digits[digit_index--] = string_nibbleToHex(value & 0x0F);
      value = value >> 4;
      }
   while ( value > 0 );

   for ( digit_index++; digit_index<sizeof(digits); digit_index++)
      *destP++ = digits[digit_index];

   return append ? destP : toP;
   }


/**
 * Convert a int16 value, (it can be with sign) to a decimal representation
 * @param dest pointer to char array where to write the result must be at least 5 digits long
 * @param value the value to convert
 * @return an updated pointer to buffer where you can insert further data
 */
char *string_int16ToDecimal(char *dest, int16 value)
   {
   uint8 digits[7];
   uint8 digit_index = 6;

   if ( value == 0 )
      {
      *dest++ = '0';
      return dest;
      }

   memset(digits,0,sizeof(digits));

   if ( value < 0 )
      {
      *dest++ = '-';
      value = -value;
      }

   while ( value > 0 )
      {
      short resto = value / 10;
      unsigned char digit = value - resto * 10;
      digits[digit_index--] = '0'+digit;
      value = resto;
      }

   for ( digit_index++; digit_index<sizeof(digits); digit_index++)
      *dest++ = digits[digit_index];

   return dest;
   }

char *string_uint16ToString(char *dest, uint16 value)
   {
   unsigned char digits[10];
   unsigned char digit_index = 9;

   if ( value == 0 )
      {
      *dest++ = '0';
      return dest;
      }

   memset(digits,0,sizeof(digits));

   while ( value > 0 )
      {
      uint16 resto = value / 10;
      uint8 digit = value - resto * 10;
      digits[digit_index--] = '0'+digit;
      value = resto;
      }

   for ( digit_index++; digit_index<sizeof(digits); digit_index++)
      *dest++ = digits[digit_index];

   return dest;
   }


/**
 * Use this one to strip prefixes and matching from parameters strings
 * @param source the string to look at
 * @param match the matching string
 * @return a pointer right after the match is found, NULL otherwise or if invalid parameters
 */
char *string_starts_with ( char *source, char *match )
   {
   int index, match_len;

   if ( source == NULL || match == NULL ) return NULL;

   match_len = strlen(match);
   for (index=0; index<match_len; index++ )
     {
     if ( source[index] != match[index] ) return NULL;
     }

   return source+index;
   }


/**
 * Use this one to match suffixes
 * @param source the string to look at
 * @param match the matching string
 * @return source if the match is found, NULL otherwise or if invalid parameters
 */
char *string_ends_with ( char *source, char *match )
   {
   int index, match_len, source_len;

   if ( source == NULL || match == NULL ) return NULL;

   match_len  = strlen(match);
   source_len = strlen(source);

   // there is no point to give an empty match
   if ( match_len <= 0 ) return NULL;

   // cannot surely have a match if the source is less than the match
   if ( source_len < match_len ) return NULL;

   // I should start looking for match at the end of string
   char *look = source + (source_len - match_len);

   for (index=0; index<match_len; index++ )
     {
     if ( look[index] != match[index] ) return NULL;
     }

  return source;
   }

/**
 * returns a pointer updated to any stripped spaces
 * if NO spaces are present then the original pointer is returned
 * NOTE: It IS possible that after the stripping the string is an empty string
 * @return an updated pointer or NULL if source was NULL
 */
char * string_strip_leading_space ( char *source )
  {
  uint8 ach;

  if ( source == NULL ) return NULL;

  while ( (ach = *source) )
    {
    if ( ach != ' ') return source;

    source++;
    }

  return source;
  }

/**
 * returns a pointer updated to any stripped cr or newlines
 * cr and newlines are replaced with the value 0
 * NOTE: It IS possible that after the stripping the string is an empty string
 * @return an updated pointer or NULL if source was NULL
 */
char * string_strip_crlf ( char *source )
  {
  uint8 ach;

  if ( source == NULL ) return NULL;

  while ( (ach = *source) )
    {
    if ( ach == '\n' || ach == '\r')
       *source = 0;

    source++;
    }

  return source;
  }


/**
 * Given a nibble, that ia a number between 0 and 15 included returns the equivalente ascii hex rapresentation
 * @return '0' to 'F' if it fits or '?' if the param is invalid
 */
uint8 string_nibbleToHex ( uint8 digit )
  {
  if ( digit <= 9 ) return '0'+digit;

  // at this point I know for sure that digit is at least 10 or more
  digit -= 10;

  if ( digit <= 5 ) return 'A'+digit;

  return '?';
  }

/**
 * Writes in the given destination the hexadecimal ascii of the given value
 * @param  destP a NON null pointer with at least two bytes free
 * @param  value the value to convert
 * @return the updated pointer for subsequent use
 */
char *string_uint8ToHex ( char* destP, uint8 value)
  {
  *destP++ = string_nibbleToHex (value >> 4);
  *destP++ = string_nibbleToHex (value & 0x0F);

  return destP;
  }

/**
 * Convert the given array of bytes into the equivalent hexadecimal version
 * it ALSO includes a terminating 0 code, so you can safely use a printf
 * @param destP where to store the risul
 * @param dest_len how much space in the risul, including the zero at the end
 * @param srcP from where to get the bytes to convert
 * @param src_len how long is the buffer to convert
 * @return where you could possibly insert some more ascii or NULL if the result will NOT fit
 */
char *uint8_array_to_hex_string (char *toP, uint32 dest_len, uint8 *srcP, uint32 src_len)
   {
   if ( ! toP ) return NULL;

   if ( ! srcP ) return toP;

   // this is the total len I need, two char each byte and a zero at the end
   int need_len = src_len * 2 + 1;

   if ( dest_len < need_len )
      return NULL;

   while ( src_len-- > 0 )
      toP = string_uint8ToHex(toP, *srcP++);

   *toP = 0;

   return toP;
   }

/**
 * this is like uint8_array_to_hex_string BUT returns the given toP
 */
char *string_uint8a_to_hstring (char *const toP, uint32 dest_len, uint8 *srcP, uint32 src_len)
   {
   uint8_array_to_hex_string(toP, dest_len, srcP, src_len);
   return toP;
   }


/**
 * convert a string of hex chars into the bytes array
 * do not insert/convert more than array_len chars
 * @return the number of converted bytes
 */
int string_hex_to_bytes ( uint8 *risul_array, int array_len, char *hex_srcP )
   {
   int ins_index=0;
   uint8 ach, tmp;
   uint8 nibble_count=0;
   uint8 aval=0;

   while ( (ach=*hex_srcP++) )
      {
      if ( ins_index >= array_len ) break;

      tmp = string_DigitToNibble (ach);

      if ( nibble_count++ )
         {
         // add the lsb nibble and store it away
         risul_array[ins_index++] = aval | tmp;
         nibble_count=0;
         }
      else
         {
         // this is the msb nibble, store it away
         aval = tmp << 4;
         }
      }

   return ins_index;
   }




/**
 * Given the ascii char representing anibble return its value
 * If an invalid value is given, not in '0' -> 'F' then zero is returned
 */
uint8 string_DigitToNibble ( uint8 ch )
  {
  switch ( ch )
    {
    case '0' : return 0;
    case '1' : return 1;
    case '2' : return 2;
    case '3' : return 3;
    case '4' : return 4;
    case '5' : return 5;
    case '6' : return 6;
    case '7' : return 7;
    case '8' : return 8;
    case '9' : return 9;
    case 'A' : return 10;
    case 'a' : return 10;
    case 'B' : return 11;
    case 'b' : return 11;
    case 'C' : return 12;
    case 'c' : return 12;
    case 'D' : return 13;
    case 'd' : return 13;
    case 'E' : return 14;
    case 'e' : return 14;
    case 'F' : return 15;
    case 'f' : return 15;
    default: return 0;
    }
  }


uint8 string_TwoHexDigitToUint8 ( uint8 msb, uint8 lsb )
   {
   uint8 abyte = string_DigitToNibble(msb);
   abyte = abyte << 4 | string_DigitToNibble(lsb);
   return abyte;
   }

/**
 * This function looks for quotes in the given string and strip them by advancing the input string of one
 * and inserting a zero at the end of the string
 * WARNING: The string is NOT copied !!!! the zero insertion is done IN the passed string
 * @param inP a pointer to a 0 terminated string
 * @return a pointer to a stripped string or inP if nothing to strip
 */
char *string_strip_quotes(char *inP)
   {
	int len, quoteindex;

	if ( inP == NULL ) return NULL;

   len = strlen (inP);

   // a string of LESS than two chars cannot be quoted
   if ( len < 2 ) return inP;

   if ( inP[0] != '"') return inP;

   quoteindex = len -1;

   if ( inP[quoteindex] != '"') return inP;

   // yes I have found a quoted string, strip it out
   inP[quoteindex] = 0;

   return inP+1;
   }



char *string_append(struct Sappend * const saP, char *stringP )
   {
   if ( ! saP ) return "null";

   if ( ! stringP ) return "(null)";

   if ( ! saP->stringP ) return "[null]";

   char * const initialP = saP->stringP;
   saP->stringP = strncpyda(saP->stringP,stringP,saP->space_left);

   uint32 delta = saP->stringP-initialP;

   if ( delta > saP->space_left )
      return "?????";

   saP->space_left -= delta;

   return saP->stringP;
   }








